import click

class CLI(click.MultiCommand):
    """
    Helper of learn module.
    """
    commands = {'train': None, 'dataset': None, 'main': None}

    def __init__(self):
        super(__class__, self).__init__(name='tf', help=__class__.__doc__)

    def list_commands(self, ctx):
        return sorted(self.commands.keys())

    def get_command(self, ctx, name):
        from . import commands
        if name in self.commands and self.commands[name] is None:
            self.commands[name] = getattr(commands, name)
        return self.commands.get(name)


main = CLI()



import json
import unittest

from dxpy.task.model import task
from dxpy.time.timestamps import TaskStamp
from dxpy.time.utils import strp


class TestTask(unittest.TestCase):
    def test_to_json(self):
        t = task.Task(tid=10, desc='test', workdir='/tmp/test',
                      worker=task.Worker.MultiThreading,
                      ttype=task.Type.Regular,
                      dependency=[1, 2, 3],
                      time_stamp=TaskStamp(create=strp(
                          "2017-09-22 12:57:44.036185")),
                      data={'sample': 42},
                      is_root=True)
        s = t.to_json()
        dct = json.loads(s)
        self.assertEqual(dct['id'], 10)
        self.assertEqual(dct['desc'], 'test')
        self.assertEqual(dct['dependency'], [1, 2, 3])
        self.assertEqual(dct['data'], {'sample': 42})
        self.assertEqual(dct['type'], 'Regular')
        self.assertEqual(dct['workdir'], '/tmp/test')
        self.assertEqual(dct['worker'], 'MultiThreading')
        self.assertEqual(dct['is_root'], True)
        self.assertEqual(dct['time_stamp'], {
                         'create': "2017-09-22 12:57:44.036185", 'start': None, 'end': None})
        self.assertEqual(dct['state'], 'BeforeSubmit')

    def test_from_json(self):
        dct = {
            '__task__': True,
            'id': 10,
            'desc': 'test',
            'workdir': '/tmp/test',
            'worker': 'Slurm',
            'type': 'Script',
            'dependency': [1, 2, 3],
            'data': {'sample': 42},
            'is_root': True,
            'time_stamp': {
                'create': "2017-09-22 12:57:44.036185",
                'start': None,
                'end': None
            },
            'state': 'BeforeSubmit'
        }
        t = task.Task.from_json(json.dumps(dct))
        self.assertEqual(t.id, 10)
        self.assertEqual(t.desc, 'test')
        self.assertEqual(t.workdir, '/tmp/test')
        self.assertEqual(t.worker, task.Worker.Slurm)
        self.assertEqual(t.type, task.Type.Script)
        self.assertEqual(t.dependency, [1, 2, 3])
        self.assertEqual(t.data, {'sample': 42})
        self.assertEqual(t.is_root, True)
        self.assertEqual(t.time_stamp.create, strp(
            "2017-09-22 12:57:44.036185"))
        self.assertEqual(t.state, task.State.BeforeSubmit)

    def test_submit(self):
        t = task.Task(10, 'test', state=task.State.BeforeSubmit)
        self.assertEqual(t.state, task.State.BeforeSubmit)
        t = task.submit(t)
        self.assertEqual(t.state, task.State.Pending)

    def test_start(self):
        t = task.Task(10, 'test', state=task.State.BeforeSubmit)
        self.assertEqual(t.state, task.State.BeforeSubmit)
        t = task.start(t)
        self.assertEqual(t.state, task.State.Runing)

    def test_complete(self):
        t = task.Task(10, 'test', state=task.State.BeforeSubmit)
        self.assertEqual(t.state, task.State.BeforeSubmit)
        t = task.complete(t)
        self.assertEqual(t.state, task.State.Complete)



def device_name(device_type, device_id=None):
    if device_id is None:
        device_id = 0
    return '/{t}:{i}'.format(t=device_type, i=device_id)


def refined_tensor_or_graph_name(tensor_or_graph):
    import tensorflow as tf
    if isinstance(tensor_or_graph, tf.Tensor):
        return tensor_or_graph.name.replace(':', '_')
    return tensor_or_graph.name

def pre_work(device=None):
    import tensorflow as tf
    from dxpy.learn.scalar import create_global_scalars
    if device is None:
        create_global_scalars()
    else:
        with tf.device(device):
            create_global_scalars()

def load_yaml_config(filename):
    from ..config import config
    import yaml
    with open(filename) as fin:
        cfg = yaml.load(fin)
    config.update(cfg)
    if 'include' in cfg:
        for f in cfg['include']:
            with open(f) as fin:
                cfg_i = yaml.load(fin)
                config.update(cfg_i)



import sys


class ExceptionHook:
    instance = None

    def __call__(self, *args, **kwargs):
        if self.instance is None:
            from IPython.core import ultratb
            self.instance = ultratb.FormattedTB(mode='Plain',
                                                color_scheme='Linux', call_pdb=1)
        return self.instance(*args, **kwargs)


def enter_debug():
    sys.excepthook = ExceptionHook()


_last_msg_file = None

dbgmsg_switch = None

def dbgmsg(*args, **kwargs):
    global _last_msg_file
    global dbgmsg_switch
    if dbgmsg_switch is False:
        return
    from .config import verbose
    if verbose <= 0:
        return
    from inspect import currentframe, getframeinfo
    frameinfo = getframeinfo(currentframe().f_back)
    fn = frameinfo.filename
    if _last_msg_file is None or not _last_msg_file == fn:
        _last_msg_file = fn
        print('[DEBUG] MESSAGE(S) FROM:', _last_msg_file)
    print('[DEBUG:{ln}]'.format(ln=frameinfo.lineno), *args, **kwargs)



from flask_restful import Resource
from .model import Path


class PathResource(Resource):
    def get(self, path):
        return {'path': Path(path).abs}


def add_api(api, root, name='path'):
    api.add_resource(PathResource, root + '/{0}/<path>'.format(name))



from .exceptions import ServiceNotFound

services = dict()


def get_default_service(key):
    if key == 'config':
        from . import configs as config_service
        return config_service
    elif key == 'database':
        c = get_or_create_service('config').get_config('database')
        if c.use_web_api:
            from .database import api as service
        else:
            from .database.service import Service as service
        return service
    raise ServiceNotFound(key)


def create_service(key, service=None):
    if service is None:
        service = get_default_service(key)
    services[key] = service

# TODO: add support for start/restart/stop service


def get_or_create_service(key):
    if key not in services:
        create_service(key)
    return services[key]



from dxpy.configs import configurable
from dxpy.learn.config import config
from dxpy.core.path import Path


@configurable(config, with_name=True)
def get_dataset(name, dataset_cls_name, dataset_name=None):
    """
    Args:
        name: config name
    """
    if dataset_name is None:
        dataset_name = name
    if dataset_cls_name == 'sraps':
        from .zoo.sraps import AnalyticalPhantomSinogramDatasetForSuperResolution 
        return AnalyticalPhantomSinogramDatasetForSuperResolution(name=dataset_name)
    elif dataset_cls_name == 'sin':
        from .zoo.sin import SinDataset
        return SinDataset(name=dataset_name)        
    else:
        raise ValueError(
            'Unknown dataset_name (class name) {}.'.format(dataset_cls_name))



"""
Utilities to measure progress of program.
"""

import time


class ProgressTimer:
    def __init__(self, nb_steps=100, min_elp=1.0):
        self._nb_steps = nb_steps
        self._step = 0
        self._start = None
        self._elaps = None
        self._pre = None
        self._min_elp = min_elp
        self.reset()

    def reset(self):
        self._start = time.time()
        self._pre = 0.0
        self._step = 0

    def event(self, step=None, msg='None'):
        if step is None:
            step = self._step
            self._step += 1
        else:
            self._step = step

        self._elaps = time.time() - self._start
        if self._elaps - self._pre < self._min_elp:
            return
        comp_percen = float(step) / float(self._nb_steps)
        if comp_percen > 0:
            eta = (1 - comp_percen) * self._elaps / comp_percen
        else:
            eta = None

        time_pas = str(datetime.timedelta(seconds=int(self._elaps)))
        time_int_v = self._elaps / (step + 1.0)
        if time_int_v < 60:
            time_int_msg = '%0.2fs/it' % (time_int_v)
        else:
            time_int_msg = str(datetime.timedelta(
                seconds=int(time_int_v)))
        if eta is None:
            time_eta = 'UKN'
        else:
            time_eta = str(datetime.timedelta(seconds=int(eta)))
        click.echo("i=%6d, %s, [%s<%s] :" %
                   (step, time_int_msg, time_pas, time_eta) + msg)
        self._pre = self._elaps



"""
Task templates of frequently used ones.
"""
import rx
import os
from dxpy.filesystem import Path
from dxpy.graph.depens import DenpensGraph
from dxpy.exceptions.checks import assert_same_length
from . import task


def task_graph(tasks, depens):
    assert_same_length((tasks, depens), ('tasks', 'depens'))
    depens_tasks = []
    for i, ds in enumerate(depens):
        if ds is None:
            depens_tasks.append([None])
        elif isinstance(ds, int):
            depens_tasks.append([tasks[ds]])
        else:
            depens_tasks.append([tasks[d] for d in ds])
    g = DenpensGraph(tasks, depens_tasks)
    for t in g.nodes():
        t.is_root = g.is_root(t)
    return g


def task_command(command, worker=None, *args, **kwargs):
    if worker is None:
        worker = task.Worker.MultiThreading
    return task.Task(*args, **kwargs, worker=worker, ttype=task.Type.Command, data={
        'command': command
    })


def task_script(file, *args, **kwargs):
    return task.Task(*args, **kwargs, ttype=task.Type.Script, data={
        'file': (Path(kwargs['workdir']) / file).abs
    })


def task_slurm(file, *args, workdir, **kwargs):
    return task.Task(*args, **kwargs, workdir=workdir, worker=task.Worker.Slurm, ttype=task.Type.Script, data={
        'file': (Path(workdir) / file).abs
    })



from dxpy.file_system.path import Path
from dxpy.task.representation import creators, task
from dxpy.task import interface

def quick_command(command):
    from fs.tempfs import TempFS
    with TempFS(auto_clean=False) as t:
        t = creators.task_command(command, workdir=t.getsyspath('.'))
        interface.create(t)




""" Base class definition """
import tensorflow as tf
from contextlib import contextmanager
from dxpy.core.path import Path
from dxpy.collections.dicts import DXDict
from typing import Dict

# On restrict mode, tensors as allocated to CPU memory if not specified.
RESTRICT_MODE = True


class GraphsManager:
    _graphs = dict()

    @classmethod
    def get_graph(cls, name):
        return cls._graphs.get(name)

    # @classmethod
    # def get_name(cls, name):
    #     import re
    #     while name in _graphs:
    #         base = name.name
    #         p = re.compile("(\w+)_(\d+)")
    #         mat = p.match(base)
    #         if mat:
    #             prefix = mat[1]
    #             gid = mat[2]
    #             base = "{}_{}".format(prefix, gid + 1)
    #         else:
    #             base = base + '_1'
    #         return name.father_path / base

    @classmethod
    def register(cls, graph):
        if graph.name in cls._graphs:
            raise ValueError(
                "Graph with name {} already registed".format(str(graph.name)))


class NodeKeys:
    """
    Pre-defined frequently used keys
    """
    EVALUATE = 'evaluate'
    INFERENCE = 'inference'
    TRAINER = 'trainer'
    SAVER = 'saver'
    LOSS = 'loss'
    MAIN = 'main'
    INPUT = 'input'
    LABEL = 'label'
    OUTPUT = 'output'
    MAIN_MODEL = 'main_model'
    CHILD_MODEL = 'child_model'
    TASK = 'task'


class Graph:
    """
    Base class of components.

    A `Graph` is an dict like collections of nodes and edges. Nodes of `Graph` can be a tensor another sub `Graph`.
    Edges representing relations of sub graphs/tensors, flow of information.

    A `Graph` is an generalization of `tf.Graph`, which is designed for following features:
        1. An unified interface of `tf.Graph` and general compute graph or operations/procedures;
        2. Seperate config and implementation, use TreeDict for configs, and supports multiple ways of config;
        3. An easy-to-use way of seperate/reuse subgraphs;
        4. Supports an warp of sessions.run/normal python function.
            Please add member method for tasks, and register them to tasks

    Config module:
    Configs module is designed to support externel hierachy configs, thus with name of graph,
    graph can load/search configs. This is designed to reduce complecity of config complex networks.
    To achieve this, use more configurable than directly pass arguments to init functions.
    For sub-models, in most cases do not pass arguments by its parent graph, directly use config system.
    There are some exceptions to this design, for some simple/frequently used blocks, some simple arguments
    may be passed to simplify config procedure. Another simulation is the case with lots of child models and
    they all share simple but same configurations, you may pass the shared arguments by its parent Graph.

    In most cases, Graph should communicate(connect) with others via dict of Tensors/Graph.


    Methods:

    -   as_tensor(self):
        return self.nodes['main'], which is designed for sub_graphs.

    -   get_feed_dict(self, task=None):
        A method which returns a feed_dict, which can be used to update parent (in most cases, the graph which called 
        subgraph.get_feed_dict()) graph's get_feed_dict() or run task.
        Which is used to garantee output nodes (if is Tensor) to be valid under certain tasks, if task is None,
        a feed_dict should be provided so that all nodes are valid.

    -   run(self, task_name, feeds):
        Call a registered function. # TODO: make it different from directly call function, provide default feeds / unified feeds, etc.

    Properties:
        name: Path, name is used for:
            1. loading configs from global config object;
            2. its basename sometimes used for namescope/variable name in TensorFlow;
    """

    def __init__(self, name, **config):
        self.name = Path(name)
        self.c = self.__load_config(config)
        self.nodes = dict()
        self.tasks = dict()

    # Methods to be overrided:
    @classmethod
    def _default_config(cls):
        """ Override this method to provide default configs. """
        return dict()

    def __hash__(self):
        return self.name.__hash__()

    def keys(self):
        return self.nodes.keys()

    def values(self):
        return self.nodes.values()

    def items(self):
        return self.nodes.items()

    def __iter__(self):
        return self.nodes.__iter__()

    def _externel_feeds(self):
        return dict()

    def get_feed_dict(self, feeds=None, task=None):
        """
        Return feed dict for this graph to work for specific tasks.

        In most cases, it works as an translator for feeds to feed_dict,
        thus replacing name of feeds to tf.Tensor.
        """
        from dxpy.collections.dicts import combine_dicts
        result = self._externel_feeds()
        if feeds is None:
            return result
        for n in self.nodes:
            if n in feeds:
                result[self.tensor(n)] = feeds[n]
            if self.nodes[n] in feeds:
                result[self.tensor(n)] = feeds[self.nodes[n]]
        return result

    def _print_config_kernel(self, recursive, indent, fout):
        title = "{ind}>{cls}:{name}({fullname})".format(ind=" " * indent,
                                                        cls=__class__,
                                                        name=self.parse_name())
        indent_sub = indent + 4
        for k in self.nodes:
            if isinstance(self.nodes[k], tf.Tensor):
                print('{ind}tf.Tensor:{name}({sp})'.format(ind=" " * indent_sub,
                                                           name=k, sp=self.nodes[k].shape), file=fout)
            elif isinstance(self.nodes[k], Graph):
                if recursive:
                    self.nodes[k].print_config(recursive, fout, indent_sub)
                else:
                    print('{ind}Graph:{name}({sub_name})'.format(ind=" " * indent_sub,
                                                                 name=k, sub_name=self.nodes[k].name), file=fout)

    @property
    def basename(self):
        """
        Get the base name of graph name. Useful for variable_scope or name_scope of graph.
        """
        return self.name.basename

    def register_node(self, name=None, tensor_or_subgraph=None):
        from .utils import refined_tensor_or_graph_name
        if tensor_or_subgraph is None:
            tensor_or_subgraph = name
            name = refined_tensor_or_graph_name(tensor_or_subgraph)
        self.nodes[name] = tensor_or_subgraph

    def register_task(self, name, func):
        self.tasks[name] = func

    def register_main_node(self, tensor_or_subgraph):
        self.register_node(NodeKeys.MAIN, tensor_or_subgraph)

    def register_main_task(self, func):
        self.register_task(NodeKeys.MAIN, func)

    def create_variable_node(self, dtype, shape, name, *, trainable=False, init_value=None):
        if init_value is not None:
            initer = tf.constant_initializer(init_value)
        else:
            initer = None
        self.register_node(name, tf.get_variable(
            name, shape, dtype, trainable=trainable, initializer=initer))
        return self.nodes[name]

    def create_placeholder_node(self, dtype, shape, name):
        self.register_node(name, tf.placeholder(dtype, shape, name))
        return self.nodes[name]

    def create_sub_graph_node(self, name, create_func):
        self.register_node(name, create_func())
        return self.nodes[name]

    def param(self, key, feeds=None, *,  default=None, raise_key_error=True):
        """
        Best practice: always use param instead of directly using self.c
        """
        if isinstance(feeds, dict) and key in feeds:
            return feeds[key]
        result = self.c.get(key, default)
        if result is None and raise_key_error:
            raise KeyError(key)
        return result

    def tensor(self, key=None):
        if key is None:
            return self.as_tensor()
        if not key in self.nodes:
            if isinstance(key, tf.Tensor):
                for n in self.nodes:
                    if self.nodes[n] == key:
                        return self.nodes[n]
        if isinstance(self.nodes[key], tf.Tensor):
            return self.nodes[key]
        elif isinstance(self.nodes[key], Graph):
            return self.nodes[key].as_tensor()
        raise KeyError("Key {} not found in graph {}.".format(key, self.name))

    def graph(self, key=None):
        if key is None:
            key = NodeKeys.MAIN
        if key in self.nodes:
            if isinstance(self.nodes[key], Graph):
                return self.nodes[key]
            else:
                raise TypeError("Type of  self.nodes[{}] of in graph {} is {}, required Graph.".format(
                    key, self.name, type(self.nodes[key])))
        raise KeyError("Key {} not found in graph {}.".format(key, self.name))

    def print_config(self, fout=None, recursive=False, indent=0):
        self._print_config_kernel(recursive, indent, fout)

    def __getitem__(self, name):
        if name in self.nodes:
            return self.nodes[name]
        elif name in self.tasks:
            return self.tasks[name]
        else:
            return None

    def __call__(self, feeds=None):
        return self.run(NodeKeys.MAIN, feeds)

    def run(self, task_name, feeds=None):
        return self.tasks[task_name](feeds)

    def as_tensor(self):
        return self.nodes.get(NodeKeys.MAIN)

    def __load_config(self, config_direct):
        from .config import config as config_global
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts(config_direct, config_global, self._default_config())



import tensorflow as tf

from ..graph import Graph
from ..scalar import ScalarVariable
from ..model.tensor import shape_as_list

from dxpy.configs import configurable
from ..config import config
from dxpy.learn.session import get_default_session

class Trainer(Graph):
    """
    Trainer which make it easier to:
        1. Unified access for single GPU and multi GPU (auto detect);
        2. Changing learning rate made easy;
        3. Training partial of variables easily (TODO)
    """

    def _check_inputs(self, loss, variables):
        def _check_tensor(l):
            if isinstance(l, tf.Tensor):
                shape = shape_as_list(l)
                if len(shape) > 1 or (len(shape) == 1 and shape[0] > 1):
                    msg = "Loss of trainer can only be scalar or list of scalar, got loss (or element of loss) of shape {}."
                    raise ValueError(msg.format(shape))
            else:
                raise TypeError(
                    "Loss (or element of loss) of trainer can only be scalar or list of scalar, not {}.".format(type(loss)))
        if loss is None:
            raise TypeError(
                "Loss of trainer can only be scalar or list of scalar, not None.")
        if isinstance(loss, (list, tuple)):
            for l in loss:
                _check_tensor(l)
        else:
            _check_tensor(loss)
    @configurable(config, with_name=True)
    def __init__(self, name=None, loss=None, variables=None, learning_rate=1e-3, simple_mode=True, **kw):
        """
        Inputs:
            loss: scalar or list of scalar (multi gpu)
            variables: list of tensors
        """
        super(__class__, self).__init__(name, learning_rate=learning_rate, simple_mode=True, **kw)
        self._check_inputs(loss, variables)
        self.loss = loss
        if variables is None:
            variables = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES)
        self.variables = variables
        with tf.variable_scope(self.basename):
            self.register_node('learning_rate', ScalarVariable('learning_rate',
                                                               init_value=self.c['learning_rate']))
            self.optimizer = self._get_optimizer()
            self.register_main_node(self._get_train_step())
            self.register_main_task(self._train)
            self.register_task('multiply_learning_rate',
                               self._multiply_learning_rate)
            self.register_task('set_learning_rate', self._set_learning_rate)

    @classmethod
    def _default_config(cls):
        return {
            'is_multi_gpu': False,
            'learning_rate': 1e-3,
            'simple_mode': True,
            'nb_gpu': 2,
        }

    def _train(self, feeds):
        sess = get_default_session()
        sess.run(self.as_tensor(), feed_dict=feeds)

    def _multiply_learning_rate(self, feeds):
        old_value = self.nodes['learning_rate'].get_value()
        new_value = old_value * self.param('ratio', feeds)
        self.nodes['learning_rate'].set_value(new_value)

    def decay_learning_rate(self, ratio):
        self.run('multiply_learning_rate', feeds={'ratio': ratio})

    def _set_learning_rate(self, feeds):
        self.nodes['learning_rate'].set_value(self.param('learning_rate'))

    def _get_optimizer(self):
        return tf.train.AdamOptimizer(self.nodes['learning_rate'].as_tensor())

    def _get_gradients(self):
        from ..utils.general import device_name
        results = []
        if isinstance(self.loss, (list, tuple)):
            for i in range(self.c['nb_gpu']):
                with tf.device(device_name('gpu', i)):
                    results.append(self.optimizer.compute_gradients(
                        self.loss[i], self.variables))
        else:
            results.append(self.optimizer.compute_gradients(
                self.loss, self.variables))
        return results

    @classmethod
    def __analysis_grad_and_vars(cls, grad_and_vars):
        nb_gpu = len(grad_and_vars)
        variables = grad_and_vars[0][1]
        grads = [g for g, _ in grad_and_vars if g is not None]
        if len(grads) > 0:
            grad = tf.add_n(grads) / nb_gpu
        else:
            grad = None
        return nb_gpu, grad, variables

    def __maybe_clip(self, grad):
        if self.c.get('with_clip', False):
            grad = tf.clip_by_value(grad,
                                    -self.c['clipv'],
                                    self.c['clipv'])
        return grad

    def _add_summary(self, average_grads):
        pass

    def _get_train_step_full(self):
        from ..scalar import global_step
        tower_gradients = self._get_gradients()
        with tf.name_scope('train_step'), tf.device('/cpu:0'):
            average_grads = []
            for grad_and_vars in zip(*tower_gradients):
                _, grad, var = self.__analysis_grad_and_vars(grad_and_vars)
                grad = self.__maybe_clip(grad)
                average_grads.append((grad, var))
            train_op = self.optimizer.apply_gradients(
                average_grads, global_step=global_step())
            self._add_summary(average_grads)
            return train_op

    def _get_train_step_simple(self):
        from ..scalar import global_step
        return self.optimizer.minimize(self.loss, global_step(), self.variables)

    def _get_train_step(self):
        if not isinstance(self.loss, (list, tuple)):
            return self._get_train_step_simple()
        else:
            return self._get_train_step_full()



import requests
import json
from functools import wraps
from flask import Flask, make_response, request, Response
from flask_restful import Resource, reqparse, Api

from ..model import task
from .. import interface
from .. import provider


class TaskResource(Resource):
    def get(self, id):
        try:
            return Response(interface.read(id).to_json(), 200, mimetype="application/json")
        except TaskNotFoundError as e:
            return str(e), 404

    def put(self, id):
        try:
            t = task.Task.from_json(request.form['task'])
            return Response(interface.update(t), 201, mimetype="application/json")
        except TaskNotFoundError as e:
            return str(e), 404

    def delete(self, id):
        try:
            return Response(interface.delete(id), 200, mimetype="application/json")
        except TaskNotFoundError as e:
            return str(e), 404


class TasksResource(Resource):
    def get(self):
        tasks = []
        interface.read_all().subscribe(lambda t: tasks.append(t))
        ts = '[\n' + ',\n'.join([t.to_json() for t in tasks]) + '\n]'
        return Response(ts, 200, mimetype="application/json")

    def post(self):
        t = task.Task.from_json(request.form['task'])
        res = interface.create(t)
        return Response(json.dumps({'id': res}), 201, mimetype="application/json")


def add_apis(api, root, name='task'):
    root_tsk = root + '/{0}'.format(name)
    api.add_resource(TaskResource, root_tsk + '/<int:id>')
    api.add_resource(TasksResource, root_tsk + 's')


def add_api(api):
    c = provider.get_or_create_service('config').get_config('interface')
    api.add_resource(TaskResource, c.task_url + '/<int:id>')
    api.add_resource(TasksResource, c.tasks_url)


def lauch_database_server():
    app = Flask(__name__)
    api = Api(app)
    add_api(api)
    c = provider.get_or_create_service('config').get_config('interface')
    app.run(host=c.ip, port=c.port, debug=c.debug)



import tensorflow as tf
import numpy as np
from dxpy.code.exceptions import DXPYTypeError
from .tensor import shape_as_list


def assert_scalar(tensor):
    if isinstance(tensor, tf.Tensor):
        shape = shape_as_list(tensor)
        if len(shape) > 1 or (len(shape) == 1 and shape[0] > 1):
            msg = "Assert scalar failed for tensor with shape{}."
            raise ValueError(msg.format(shape))



import tensorflow as tf
from .blocks import Conv2D, StackedConv2D, InceptionBlock
from ..base import Model, NodeKeys
from ..tensor import shape_as_list
from dxpy.configs import configurable
from ...config import get_configs_view

class ResidualIncept(Model):
    @configurable(get_configs_view(), with_name=True)
    def __init__(self, name, input_tensor, ratio=0.3, paths=3, **kw):
        super().__init__(name, inputs=input_tensor, ratio=ratio, paths=paths, **kw)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'ratio': 0.1,
            'paths': 3,
        }, super()._default_config())

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        h = InceptionBlock('incept', x, paths=self.param('paths'))()
        with tf.name_scope('add'):
            x = x + h * self.param('ratio')
        return x


class ResidualStackedConv(Model):
    @configurable(get_configs_view(), with_name=True)
    def __init__(self, name, input_tensor, *, nb_layers=None, block_type=None, ratio=0.1, **config):
        super().__init__(name, inputs=input_tensor, nb_layers=nb_layers, ratio=ratio, **config)



    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'ratio': 0.1,
            'nb_layers': 2,
        }, super()._default_config())

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        h = StackedConv2D('convs', x, nb_layers=self.param('nb_layers'),
                          activation='res_celu', filters=shape_as_list(x)[-1])()
        with tf.name_scope('add'):
            return x + h * self.param('ratio')


class StackedResidual(Model):
    """
    Sub model name: sub_{0..nb_layers}
    """
    INCEPT_TYPE = 'incept'
    STACKED_CONV_TYPE = 'stacked_conv'

    @configurable(get_configs_view(), with_name=True)
    def __init__(self, name, input_tensor, *, nb_layers=None, block_type=None, **config):
        super().__init__(name, inputs=input_tensor,
                         nb_layers=nb_layers, block_type=block_type, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'nb_layers': 10,
            'block_type': cls.INCEPT_TYPE,
        }, super()._default_config())

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        for i in range(self.param('nb_layers')):
            name = self.name / 'res_{}'.format(i)
            if self.param('block_type') == self.INCEPT_TYPE:
                x = ResidualIncept(name, x)()
            elif self.param('block_type') == self.STACKED_CONV_TYPE:
                x = ResidualStackedConv(name, x)()
        return x

from dxpy.configs import configurable
from ...config import config
class StackedResidualv2(Model):
    """
    Sub model name: sub_{0..nb_layers}
    """
    INCEPT_TYPE = 'incept'
    STACKED_CONV_TYPE = 'stacked_conv'
    @configurable(config, with_name=True)
    def __init__(self, name, input_tensor, *, nb_layers=10, block_type='incept', **kw):
        super().__init__(name, inputs=input_tensor,
                         nb_layers=nb_layers, block_type=block_type, **kw)

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        for i in range(self.param('nb_layers')):
            name = self.name / 'res_{}'.format(i)
            if self.param('block_type') == self.INCEPT_TYPE:
                x = ResidualIncept(name, x)()
            elif self.param('block_type') == self.STACKED_CONV_TYPE:
                x = ResidualStackedConv(name, x)()
        return x



import click


@click.group()
def ui():
    pass


@ui.command()
def start():
    from ..web import lauch_database_server
    lauch_database_server()


from ...database import cli as db
from ...run.cli import run



default_engine = None


class MatlabEngine:
    def __init__(self):
        self.eng = None
        self.pre = None

    def __enter__(self):
        global default_engine
        import matlab.engine
        import os
        from dxpy.core.path import Path
        self.eng = matlab.engine.start_matlab()
        path_dxmat = os.environ['PATH_DXL_DXMAT']
        path_gen = Path(path_dxmat) / 'phantom'
        self.eng.addpath(str(path_gen))
        self.pre = default_engine
        default_engine = self.eng
        return self.eng

    def __exit__(self, type, value, trackback):
        global default_engine
        default_engine = self.pre
        self.eng.quit()

    @staticmethod
    def get_default_engine():
        return default_engine


def call_matlab_api(func):
    if MatlabEngine.get_default_engine() is None:
        with MatlabEngine() as eng:
            return func(eng)
    else:
        return func(MatlabEngine.get_default_engine())



import tensorflow as tf
from ..base import Model, NodeKeys, ModelPipe
from .. import activation
from ..tensor import shape_as_list

from dxpy.configs import configurable
from ...config import get_configs_view

class Conv2D(Model):
    def __init__(self, name='conv2d', input_tensor=None, filters=None, kernel_size=None,
                 *,
                 strides=None, padding=None, activation=None, **config):
        super().__init__(name, input_tensor,
                         filters=filters,
                         kernel_size=kernel_size,
                         strides=strides,
                         padding=padding,
                         activation=activation,
                         **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'kernel_size': 3,
            'strides': (1, 1),
            'padding': 'same',
            'activation': 'basic'
        }, super()._default_config())

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        acc = activation.unified_config(self.param('activation', feeds))
        x = activation.apply(acc, x, 'pre')
        x = tf.layers.conv2d(x, self.param('filters', feeds),
                             self.param('kernel_size', feeds),
                             self.param('strides', feeds),
                             self.param('padding', feeds), name='convolution')
        x = activation.apply(acc, x, 'post')
        return x


class StackedConv2D(Model):
    @configurable(get_configs_view(), with_name=True)
    def __init__(self, name, input_tensor, *,
                 nb_layers=None,
                 activation=None,
                 filters=None, **config):
        super().__init__(name, input_tensor, nb_layers=nb_layers,
                         filters=filters, activation=activation, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        cfg = {
            'activation': 'basic',
            'padding': 'same',
            'filters': 32
        }
        return combine_dicts(cfg, super()._default_config())

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        for i in range(self.param('nb_layers')):
            x = Conv2D('conv2d_{}'.format(i), x, self.param('filters'),
                       activation=self.param('activation'),
                       padding=self.param('padding'))()
        return x


class InceptionBlock(Model):
    def __init__(self, name='incept', input_tensor=None, activation=None, **config):
        super().__init__(name, inputs={NodeKeys.INPUT: input_tensor}, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'paths': 3,
            'activation': 'incept',
        }, super()._default_config())

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        filters = x.shape.as_list()[-1]
        acc = activation.unified_config(self.param('activation', feeds))
        x = activation.apply(acc, x, 'pre')
        paths = []
        for i_path in range(self.param('paths')):
            with tf.variable_scope('path_{}'.format(i_path)):
                h = Conv2D('conv2d_0', x, filters, 1,
                           activation='linear').as_tensor()
                for j in range(i_path):
                    h = Conv2D('conv2d_{}'.format(j + 1), h, filters,
                               3, activation='pre').as_tensor()
                paths.append(h)
        with tf.name_scope('concat'):
            x = tf.concat(paths, axis=-1)
        x = Conv2D('conv_end', x, filters, 1, activation='pre').as_tensor()
        return x


# class ResidualIncept(Model):
#     def __init__(self, name, input_tensor, **config):
#         super().__init__(name, inputs=input_tensor, **config)

#     @classmethod
#     def _default_config(cls):
#         from dxpy.collections.dicts import combine_dicts
#         return combine_dicts({
#             'ratio': 0.1,
#             'paths': 3,
#         }, cls._default_config())

#     def _kernel(self, feeds):
#         x = feeds[NodeKeys.INPUT]
#         h = InceptionBlock('incept', x, paths=self.param('paths'))
#         with tf.name_scope('add'):
#             x = x + h * self.param('ratio')
#         return x


# class ResidualStackedConv(Model):
#     def __init__(self, name, input_tensor, *, nb_layers=None, ratio=None, **config):
#         super().__init__(name, inputs=input_tensor,
#                          nb_layers=nb_layers, ratio=ratio, **config)

#     @classmethod
#     def _default_config(cls):
#         from dxpy.collections.dicts import combine_dicts
#         return combine_dicts({
#             'ratio': 0.1,
#             'nb_layers': 2,
#         }, super()._default_config())

#     def _kernel(self, feeds):
#         x = feeds[NodeKeys.INPUT]
#         h = StackedConv2D('convs', x, nb_layers=self.param('nb_layers'),
#                           activation='res_celu', filters=shape_as_list(x)[-1])()
#         with tf.name_scope('add'):
#             return x + h * self.param('ratio')


# class StackedResidual(Model):
#     """
#     Sub model name: sub_{0..nb_layers}
#     """

#     def __init__(self, name, input_tensor, *, nb_layers=None, block_type=None, **config):
#         super().__init__(name, inputs=input_tensor, nb_layers=nb_layers, **config)

#     @classmethod
#     def _default_config(cls):
#         from dxpy.collections.dicts import combine_dicts
#         return combine_dicts({
#             'nb_layers': 10,
#             'block_type': 'incept',
#         }, cls._default_config())

#     def _kernel(self, feeds):
#         x = feeds[NodeKeys.INPUT]
#         for i in range(self.param('nb_layers')):
#             name = self.name / 'res_{}'.format(i)
#             if self.param('block_type') == 'incept':
#                 x = ResidualIncept(name, x)
#             elif self.param('block_type') == 'stacked_conv':
#                 x = ResidualStackedConv(name, x)
#         return x



import numpy as np


def __guess_shape(length, ndim=None):
    if length == 1 and (ndim is None or ndim == 0):
        return []
    if ndim is None:
        return [length]
    size = int(np.ceil(np.power(length, 1 / ndim)))
    if size ** ndim == length:
        return [size] * ndim
    raise ValueError(
        "Cannot guess shape of length {} with ndim {}.".format(length, ndim))


def load_bin(filename, shape=None, dtype=np.float32, *, fs=None, ndim=None):
    open_func = open if fs is None else fs.open
    with open_func(filename, 'rb') as fin:
        raw_data = np.fromstring(fin.read(), dtype)
    if shape is None:
        shape = __guess_shape(raw_data.size, ndim)
    return np.reshape(raw_data, shape)



from dxpy.learn.graph import Graph
from dxpy.configs import configurable
from dxpy.learn.config import config
import tensorflow as tf
import numpy as np

class SinDataset(Graph):
    @configurable(config, with_name=True)
    def __init__(self, name, batch_size=32, amplitude=1.0, frequency=1.0, phase=0.0):
        super().__init__(name=name, batch_size=batch_size, amplitude=amplitude, frequency=frequency, phase=phase)
        self._construct_dataset()

    def _construct_dataset(self):
        from dxpy.learn.utils.tensor import ensure_shape
        with tf.name_scope(str(self.name)):
            x = tf.random_uniform(shape=[self.param('batch_size'), 1], minval=0, maxval=2*np.pi)
            y = tf.sin(x*self.param('frequency')+self.param('phase'))*self.param('amplitude')
            # dataset = tf.data.Dataset.from_tensors({'x': x, 'y': y})
            # dataset = dataset.batch(self.param('batch_size'))
            # iterator = dataset.make_one_shot_iterator()
            # next_element = iterator.get_next()
            shape = [self.param('batch_size'), 1]
            self.register_node('x', ensure_shape(x, shape=shape))
            self.register_node('y', ensure_shape(y, shape=shape))

    




from dxpy.core.path import Path


class ConfigsView:
    def __init__(self, dct, base_path=''):
        self.data = dct
        self.base = Path(base_path)

    def __unified_keys(self, path_or_keys):
        if isinstance(path_or_keys, (list, tuple)):
            return list(self.base.parts()), list(path_or_keys)
        else:
            return list(self.base.parts()), list(Path(path_or_keys).parts())

    def _get_value_raw(self, keys):
        result = self.data
        for k in keys:
            if not isinstance(result, (dict, ConfigsView)):
                return None
            result = result.get(k)
        return result

    def _form_path(self, keys):
        return '/'.join(keys)

    def _query(self, base_keys, local_keys):
        # if len(local_keys) <= 1:
        result = self._get_value_raw(base_keys + local_keys)
        # else:
            # path = self.base / local_keys[0]
            # key_path = self._form_path(local_keys[1:])
            # result = ConfigsView(self.data, path)._query(key_path)
        return result

    def _search(self, key):
        base_keys, local_keys = self.__unified_keys(key)
        result = self._query(base_keys, local_keys)
        path = self._form_path(base_keys + local_keys)
        if result is None and len(base_keys) > 0:
            new_path = self._form_path((base_keys+local_keys)[:-2])
            result, _ = ConfigsView(self.data,new_path)._search(local_keys[-1])
        return result, path

    def _post_processing(self, result, path, default, restrict):
        if isinstance(result, dict) or (result is None and default is None and not restrict):
            result = ConfigsView(self.data, path)
        elif result is None:
            result = default
        return result

    def to_dict(self):
        result, path = self._search(str(self.base))
        if isinstance(result, dict):
            return result
        else:
            raise ValueError("Configs view on base path is not dict.")

    def get(self, key, default=None, *, restrict=False):
        result, path = self._search(key)
        return self._post_processing(result, path, default, restrict)

    def __getitem__(self, key):
        return self.get(key, restrict=True)

    def __iter__(self):
        dct = self._get_value_raw(self.base.parts())
        if dct is None:
            return list().__iter__()
        else:
            return dct.__iter__()

    


def child_view(configs_view, extend_path):
    return ConfigsView(configs_view.base_path / str(extend_path), configs_view.data)



from dxpy.configs import configurable
from dxpy.learn.config import config
from dxpy.core.path import Path


@configurable(config, with_name=True)
def get_network(name, dataset, network_cls_name, network_name=None, **kw):
 
    from .base import NodeKeys
    if network_name is None:
        network_name = name
    if network_cls_name == 'srms':
        from .zoo.srms import SRMultiScale
        # valid_keys = ['input/image{}'.format(2**i) for i in range(10)]
        # valid_keys += ['label/image{}'.format(2**i) for i in range(10)]
        # inputs = dict()
        inputs = dataset
        # for k in valid_keys:
            # if k in dataset.nodes:
                # inputs[k] = dataset[k]
        return SRMultiScale(name=network_name, inputs=inputs, **kw)
    elif network_cls_name == 'sin':
        from .zoo.sin import SinNet
        return SinNet(name=name, inputs={NodeKeys.INPUT: dataset['x'], NodeKeys.LABEL: dataset['y']}, **kw)

    else:
        raise ValueError(
            'Unknown network name (class name) {}.'.format(dataset_cls_name))




import numpy as np
from dxpy.matlab import call_matlab_api


def generate_shepp_logans(nb_images, nb_size, sigma, *, backend='matlab'):
    if backend == 'matlab':
        return np.array(call_matlab_api(lambda e: e.GenerateSheppLogans(nb_images, float(nb_size), float(sigma))))
    raise ValueError("Unknown backend {}.".format(backend))


def generate_phantom(shape=256.0, *, backend='matlab'):
    if backend == 'matlab':
        return np.array(call_matlab_api(lambda eng: eng.phantom(shape)))
    raise ValueError("Unknown backend {}.".format(backend))



from dxpy.configs import configurable
import tensorflow as tf
from ..config import config
from ..model.tensor import ReduceSum


@configurable(config, with_name=True)
def poission_noise_simulator(tensor, summation_value, name='poission_noise_simulator')
    tensor = ReduceSum('{}/reduce_sum'.format(name),
                       tensor,
                       fixed_summation_value=summation_value).as_tensor()
    with tf.name_scope('add_poission_noise'):
        tensor = tf.random_poisson(tensor, shape=[])
    return tensor



from enum import Enum

import numpy as np
import tensorflow as tf

from ..base import Model, NodeKeys

DENORM_INPUT_KEY = 'to_denorm'
DENORM_OUTPUT_KEY = 'denormed'


class ConfigKeys(Enum):
    DENORM_INPUT_KEY = 'to_denorm'
    DENORM_OUTPUT_KEY = 'denormed'
    ADD_DENORM_FLAG = 'add_denorm'
    WITH_BATCH_DIM = 'with_batch_dim'


class Normalizer(Model):
    def __init__(self, name, inputs=None, **config):
        super().__init__(name, inputs, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            ConfigKeys.ADD_DENORM_FLAG.name: False,
            ConfigKeys.WITH_BATCH_DIM.name: True,
        }, super()._default_config())

    def _pre_kernel_pre_inputs(self):
        from ..tensor import PlaceHolder
        super()._pre_kernel_pre_inputs()
        if self.param(ConfigKeys.ADD_DENORM_FLAG.name) and not DENORM_INPUT_KEY in self.inputs:
            self.inputs[DENORM_INPUT_KEY] = PlaceHolder(self.inputs[NodeKeys.INPUT].shape,
                                                        self.inputs[NodeKeys.INPUT].dtype,
                                                        name=DENORM_INPUT_KEY)

    def _kernel(self, feeds):
        with tf.name_scope('normalization'):
            result = self._normalization_kernel(feeds)
        if DENORM_INPUT_KEY in feeds:
            with tf.name_scope('denormalization'):
                denorm_feeds = dict()
                if isinstance(result, dict):
                    denorm_feeds.update(result)
                else:
                    denorm_feeds.update({NodeKeys.MAIN, result})
                denorm_feeds.update(feeds)
                result.update(self._denormalization_kernel(denorm_feeds))
        return result


class FixWhite(Normalizer):
    def __init__(self, name='normalizer/fix_white', inputs=None, **config):
        super().__init__(name, inputs, **config)

    def _normalization_kernel(self, feeds):
        return (feeds[NodeKeys.INPUT] - self.param('mean', feeds)) / self.param('std', feeds)

    def _denormalization_kernel(self, feeds):
        return feeds[NodeKeys.MAIN] * self.param('std', feeds) + self.param('mean', feeds)


class SelfMinMax(Normalizer):
    def __init__(self, name='normalizer/self_min_max', inputs=None, **config):
        super().__init__(name, inputs, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'add_denorm': True
        }, super()._default_config())

    def _normalization_kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        rmin = tf.reduce_min(x)
        rmax = tf.reduce_max(x)
        data = (x - rmin) / (rmax - rmin)
        return {'min': rmin, 'max': rmax, NodeKeys.MAIN: data}

    def _denormalization_kernel(self, feeds):
        return {DENORM_OUTPUT_KEY: feeds[DENORM_INPUT_KEY] * (feeds['max'] - feeds['min']) + feeds['min']}


class SelfMeanStd(Normalizer):
    def __init__(self, name='normalizer/self_mean_std', inputs=None, **config):
        super().__init__(name, inputs, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'add_denorm': True
        }, super()._default_config())

    def _normalization_kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        mean, stdv = tf.nn.moments(x)
        return {'mean': mean, 'std': stdv, NodeKeys.MAIN: (x - mean) / stdv}

    def _denormalization_kernel(self, feeds):
        return {DENORM_OUTPUT_KEY: feeds[DENORM_INPUT_KEY] * feeds['std'] + feeds['mean']}


class ReduceSum(Normalizer):
    """
    Normalize tensor to fix summation.
    """

    def __init__(self, name='normalizer/reduce_sum',
                 inputs=None,
                 fixed_summation_value=None,
                 **config):
        super().__init__(name, inputs, fixed_summation_value=fixed_summation_value, **config)

    def _normalization_kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        if self.param(ConfigKeys.WITH_BATCH_DIM.name):
            summation = tf.reduce_sum(x, axis=list(range(1, len(x.shape))),
                                      keep_dims=True)
        else:
            summation = tf.reduce_sum(x)
        return {'sum': summation,
                NodeKeys.MAIN: x / summation * self.param('fixed_summation_value', feeds)}

    def _denormalization_kernel(self, feeds):
        return {DENORM_OUTPUT_KEY: feeds[DENORM_INPUT_KEY] * feeds['sum'] / self.param('fixed_summation_value', feeds)}


def get_normalizer(name):
    if name.lower() in ['self_min_max', 'selfminmax']:
        return SelfMinMax
    if name == 'reduce_sum':
        return ReduceSum



import tensorflow as tf
import numpy as np
from ..base import Model, NodeKeys


def resize(inputs, size, method=None, name='upsampling'):
    if method is None:
        method = 'bilinear'
    input_shape = inputs.shape.as_list()
    h0, w0 = input_shape[1:3]
    h1 = int(h0 * size[0])
    w1 = int(w0 * size[1])
    with tf.name_scope(name):
        if method == 'nearest':
            h = tf.image.resize_nearest_neighbor(inputs, size=[h1, w1])
        elif method == 'bilinear':
            h = tf.image.resize_bilinear(inputs, size=[h1, w1])
    return h


def add_poisson_noise(input_tensor):
    if not isinstance(input_tensor, tf.Tensor):
        raise TypeError("Requre Tensor, got {}.".format(type(input_tensor)))
    with tf.name_scope('add_poisson_noise'):
        if input_tensor.dtype != tf.float32:
            input_tensor = tf.cast(input_tensor, tf.float32)
        return tf.random_poisson(input_tensor, [])


class DownSampler(Model):
    def __init__(self,
                 input_tensor, ratio,
                 *,
                 method=None, padding=None, name='down_sampler',
                 **config):
        super().__init__(name, input_tensor,
                         ratio=ratio, method=method, padding=padding, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        cfg = {
            'method': 'mean',
            'padding': 'same',
            'keep_energy': False,
        }
        return combine_dicts(cfg, super()._default_config())

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]
        if self.param('method') == 'mean':
            with tf.name_scope('down_sample'):
                x = tf.nn.avg_pool(x,
                                   [1] + list(self.param('ratio')) + [1],
                                   padding=self.param('padding').upper(),
                                   strides=[1] + list(self.param('ratio')) + [1])
                if self.param('keep_energy'):
                    x = x * np.prod(self.param('ratio'))
        else:
            raise ValueError("Unsupported down sample method")
        return x


class MultiDownSampler(Model):
    """
    A helper model to create multiple down sampler and form their results to a dict.
    """

    def __init__(self, input_tensor, down_sample_ratios=None,
                 *,
                 keep_original=None,
                 original_key=None,
                 with_shape_info=None,
                 original_shape=None,
                 name='multi_down_sampler',
                 **config):
        super().__init__(name,
                         input_tensor,
                         down_sample_ratios=down_sample_ratios,
                         keep_original=keep_original,
                         original_key=original_key, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        cfg = {
            'keep_original': True,
            'original_key': 'ds1x',
            'rigister_output_with_prefix': False,
            'with_shape_info': False
        }
        return combine_dicts(cfg, super()._default_config())

    def _kernel(self, feeds):
        image_orgin = feeds[NodeKeys.INPUT]
        result = dict()
        for n, v in self.param('down_sample_ratios').items():
            result[n] = DownSampler(image_orgin, v, name=self.name / n)()
        if self.param('keep_original'):
            result[self.param('original_key')] = self.tensor(NodeKeys.INPUT)
        return result


class Padder(Model):
    def __init__(self, input_tensor, padding=None, shape=None, offset=None, name='padder', **config):
        """
        Args:
            padding: str|list of str, value|element value is one of the flollowing:
            -   none: no padding on specific axis
            -   period: period padding
            -   shape: list of shape (target shape)
            -   offset: start index of original image in result image
        """
        super().__init__(name,
                         inputs={NodeKeys.INPUT: input_tensor},
                         padding=padding, shape=shape)

    def _kernel(self, feeds):
        x = feeds[NodeKeys.INPUT]


def random_crop(input_, target_shape, name='random_crop'):
    from dxpy.tensor.transform import random_crop_offset
    from ..tensor import ensure_tensor, shape_as_list
    with tf.name_scope(name):
        input_shape = shape_as_list(input_)
        target_shape = shape_as_list(target_shape)
        random_offset = tf.py_func(random_crop_offset,
                                   [input_shape, target_shape], tf.int64)
        return tf.slice(input_, random_offset, target_shape)


def align_crop(input_, target, offset=None, name='align_crop'):
    """
    Args:
        input_: tensor to be cropped
        target: tensor to fit (provide target shape)
        offset: start crop offset of dim 1 and 2, if `None`, auto calculated to perform center crop
    """
    from ..tensor import shape_as_list
    with tf.name_scope(name):
        shape_input = shape_as_list(input_)
        shape_output = shape_as_list(target)
        if offset is None:
            offset = [0] + [(shape_input[i] - shape_output[i]) // 2
                            for i in range(1, 3)] + [0]
        return tf.slice(input_, offset, shape_output)


def boundary_crop(input_, offset, name="boundary_crop"):
    from ..tensor import shape_as_list
    with tf.name_scope(name):
        shape = shape_as_list(input_)
        if len(offset) == 2:
            offset = [0] + list(offset) + [0]
        shape_output = [s - 2 * o for s, o in zip(shape, offset)]
        return tf.slice(input_, offset, shape_output)



import json
import unittest

import rx

from dxpy.task.database.config import config as c
from dxpy.task import database as db
from dxpy.task.database.model import TaskDB

from dxpy.task.exceptions import TaskNotFoundError


class TestDataBase(unittest.TestCase):
    def setUp(self):
        c['path'] = ':memory:'
        c['use_web_api'] = False
        db.Database.create()
        self.dummy_dct = {
            '__task__': True,
            'id': 1,
            'desc': 'dummpy task',
            'data': {},
            'time_stamp': {
                'create': "2017-09-22 12:57:44.036185",
                'start': None,
                'end': None,
            },
            'is_root': True,
            'state': 'BeforeSubmit',
            'workdir': '/tmp/test',
            'worker': 'Slurm',
            'dependency': [],
            'type': 'Regular'
        }
        self.dummy_json = json.dumps(self.dummy_dct)
        self.dummy_id = db.create(self.dummy_json)
        self.modify_id = db.create(self.dummy_json)
        self.delete_id = db.create(self.dummy_json)

    def tearDown(self):
        db.Database.clear()
        c.back_to_default()

    def test_create(self):
        tid = db.create(self.dummy_json)
        self.assertIsInstance(tid, int)

    def test_creat_json(self):
        dct = {"__task__": True,
               "id": 1,
               "desc": "test",
               "workdir": "/home/hongxwing/Workspace/dxl/tests_dxpy/task",
               "worker": "NoAction",
               "type": "Regular",
               "state": "BeforeSubmit",
               "dependency": [],
               "time_stamp": {"create": "2017-10-02 05:15:56.349184", "start": None, "end": None},
               "is_root": False,
               "data": {}}
        s = json.dumps(dct)

        tid = db.create(s)

    def test_read(self):
        # TODO: add desired output json
        t = db.read(self.dummy_id)
        self.assertIsInstance(t, str)
        dct = json.loads(t)
        self.assertTrue(dct['__task__'])
        self.assertIsInstance(dct['data'], dict)

    def test_read_invalid_tid(self):
        invalid_tid = self.dummy_id + 1000
        with self.assertRaises(TaskNotFoundError):
            t = db.read(invalid_tid)

    def test_read_all(self):
        t_all = db.read_all()
        self.assertIsInstance(t_all, rx.Observable)
        t_list = (t_all
                  .subscribe_on(rx.concurrency.ThreadPoolScheduler())
                  .to_list()
                  .to_blocking().first())
        self.assertIsInstance(t_list, list)

    def test_update(self):
        new_dct = json.loads(db.read(self.modify_id))
        new_dct['desc'] = 'modified'
        db.update(json.dumps(new_dct))
        data = json.loads(db.read(self.modify_id))
        self.assertEqual(data['desc'], "modified")

    def test_delete(self):
        db.read(self.delete_id)
        db.delete(self.delete_id)
        with self.assertRaises(TaskNotFoundError):
            db.read(self.delete_id)



import h5py
import numpy as np


def save(filename, array, tensor_spec=None):    
    from .base import TensorSpec
    if tensor_spec is None:
        tensor_spec = TensorSpec()
    with h5py.File(filename, 'w') as fout:
        return tensor_spec.create_dateset_on(fout, array)


def load(tensor_spec, slices=None):
    with h5py.File(tensor_spec.filename, 'r') as fin:
        dataset = fin
        for n in tensor_spec.dataset_names[:-1]:
            dataset = dataset[n]
        if slices is None:
            return np.array(dataset[tensor_spec.dataset_names[-1]])
        else:
            return np.array(dataset[tensor_spec.dataset_names[-1]][slices])



import tensorflow as tf

# cluster = tf.train.ClusterSpec({"local": ["localhost:2222", "localhost:2223"]})
cluster = tf.train.ClusterSpec({"local": ["localhost:2222"]})
server = tf.train.Server(cluster, job_name="local", task_index=0)

from dxpy.learn.dataset.zoo.sraps import AnalyticalPhantomSinogramDatasetForSuperResolution
from dxpy.config import config
import yaml

with open('dxln.yml') as fin:
    cfg = yaml.load(fin)
config.update(cfg)

with tf.device('/job:local/task:0'):
    dataset = AnalyticalPhantomSinogramDatasetForSuperResolution()
    sinos = dataset['input/image1x']
server.join()


"""
Task database support module.

A database which react with the following JSON format data.
JSON Format:
{
    'id': int, task id;
    'desc': str, task description;
    'data': dict, any data of specific task, e.g. sid, .sh file;
    'time_stamp': dict with:
        'create': str, time of task creation
        'start': str, time of start running
        'end': str, time of complete
    'state': str, task state
    'is_root': bool, is task is a root task, i.e. which is automitically submitted by run deamon
    'worker': str, worker of action, one of 'NoAction', 'Multithreading', 'Slurm'
    'type': str, type of task, ['Regular', 'Command', 'Script']
    'workdir': str, path of workdir,
    'dependency': list, dependency ids;    
}
"""
from .api import *



"""
Abstract of path object.
Features:
    An unified model for path, pathlib.Path, fs.path, and url encoded path.
"""

import fs.path as fp
from fs.path import normpath
from fs import opener
from urllib.parse import quote_plus as qut
from urllib.parse import unquote_plus as uqut
import pathlib
from ruamel.yaml import YAML


from ..exceptions import NonePathError, ConflictUrlSpecError


class UrlSpec:
    def __init__(self, protocol=None, username=None, password=None, resource=None, params=None):
        self.protocol = protocol
        self.username = username
        self.password = password
        self.resource = resource
        self.params = params

    def __eq__(self, url_spec):
        compares = [self.protocol == url_spec.protocol,
                    self.username == url_spec.username,
                    self.password == url_spec.password,
                    self.resource == url_spec.resource]
        return all(compares)

    @classmethod
    def parse(cls, s):
        p = opener.parse.parse_fs_url(s)
        return cls(p.protocol, p.username, p.password, p.resource, p.params)

    def serilization(self, format='JSON'):
        def to_json_dct():
            return {'protocol': self.protocol,
                    'username': self.username,
                    'password': self.password,
                    'resource': self.resource,
                    'params': self.params}
        if format.upper() == 'JSON':
            pass
        return None


QUOTED_SLASH = qut('/')
DOUBLE_QUOTED_SLASH = qut(qut('/'))


def is_quoted_url(s):
    return QUOTED_SLASH in s


def is_double_quoted_url(s):
    return DOUBLE_QUOTED_SLASH in s


def url_quote_path(path):
    return qut(path)


def url_unquote_path(url):
    return uqut(url)


def url_double_quoted_path(path):
    return qut(qut(path))


def url_double_unquoted_path(path):
    return uqut(uqut(path))


def unified_path_to_str(path):
    if path is None:
        raise NotValidPathError
    if isinstance(path, str):
        if is_quoted_url(path):
            path = url_unquote_path(path)
        return normpath(path)
    if isinstance(path, Path):
        return unified_path_to_str(str(path.path))
    if isinstance(path, pathlib.Path):
        return unified_path_to_str(str(path))


class Path:
    """ Unified path for posix/windows/url/url_quoted paths.
    This class is an 'abstract' path object, which means it only normalize its representation for different platform,
    and provide functionalities of pure path calculations, without any relation to the true file system.
    Thus this class does **NOT** provide methods like `exists()` or `is_dir()`, please refer `dxpy.file_syste.file.File` for these functions.
    """
    yaml_tag = '!path'

    def __init__(self, path, url_spec=None):
        """
        """
        self.path = unified_path_to_str(path)
        self.url_spec = url_spec

    @classmethod
    def copy(cls, path, url_spec=None):
        if isinstance(path, Path):
            raw_path = fp.normpath(str(path.path))
            if url_spec is not None and not path.url_spec == url_spec:
                raise ConflictUrlSpecError
            new_url_spec = UrlSpec(url_spec.protocol, url_spec.username,
                                   url_spec.password, url_spec.resource,
                                   url_spec.params)
            return cls(raw_path, new_url_spec)
        if isinstance(path, pathlib.Path):
            return cls(str(path.absolute()), None)

    def __str__(self):
        if self.url_spec is None:
            return self.path
        else:
            fmt = "__str__ for Path with not None url_spec is not Implemented yet, path: {0}, url_spec: {1}."
            raise TypeError(fmt.format(self.path, self.url_spec))

    def __truediv__(self, name):
        return Path(fp.combine(self.abs, name), self.url_spec)

    def __add__(self, suffix):
        return Path(self.abs + suffix)

    def check_exists(self, fs):
        return fs.exists(self.abs)

    @property
    def abs(self):
        if fp.isabs(self.path):
            return self.path
        else:
            from fs.osfs import OSFS
            with OSFS('.') as fs:
                return fs.getsyspath(self.path)

    @property
    def rel(self):
        return fp.relpath(self.path)

    @property
    def isabs(self):
        return fp.isabs(self.path)

    @property
    def isrel(self):
        return not fp.isabs(self.path)

    @property
    def parts(self):
        result = fp.iteratepath(self.path)
        if fp.isabs(self.path) and not result[0] == '/':
            result = ['/'] + result
        return tuple(result)

    def rel_parts(self):
        result = fp.iteratepath(self.path)
        if result[0] == '/':
            result = result[1:]
        return result

    @property
    def recurse(self):
        return fp.recursepath(self.path)

    @property
    def father_path(self):
        return Path(fp.dirname(self.path))

    @property
    def father(self):
        return fp.dirname(self.path)

    @property
    def name(self):
        return fp.basename(self.path)

    @property
    def quoted_url(self):
        return qut(qut(self.path))

    @property
    def suffix(self):
        return fp.splitext(self.path)[1]

    def __eq__(self, path):
        return self.path == Path(path).path

    @classmethod
    def to_yaml(cls, representer, obj):
        return representer.represent_scalar(cls.yaml_tag, obj.abs)

    @classmethod
    def from_yaml(cls, constructor, node):
        return Path(constructor.construct_scalar(node))

    def __hash__(self):
        return hash(self.path)


yaml = YAML()
yaml.register_class(Path)



import datetime


def now(local=False):
    if local:
        return datetime.datetime.now()
    else:
        return datetime.datetime.utcnow()


def delta(milliseconds=0):
    return datetime.timedelta(milliseconds=milliseconds)


MS = 1
SEC = 1000 * MS
MIN = 60 * SEC
HUR = 60 * MIN
DAY = 24 * HUR


def iso_format(sep='T'):
    return "%Y-%m-%d{sep}%H:%M:%S.%f".format(sep=sep)


def strf(dt, fmt=None):
    if dt is None:
        return None
    if fmt is None:
        fmt = iso_format(' ')
    return dt.strftime(fmt)


def strp(s, fmt=None):
    if s is None:
        return None
    if fmt is None:
        fmt = iso_format(' ')
    return datetime.datetime.strptime(s, fmt)






import sys
import rx
from rx.concurrency import ThreadPoolScheduler
from .. import interface
from . import workers
from .. import provider

SUPPORTED_WORKERS = [workers.Slurm, workers.MultiThreding, workers.NoAction]


def start(task):
    workers.get_workers(task).run(task)


def read_complete_tasks_on_worker(worker) -> 'rx.Observable<Task>':
    return (interface.read_all()
            .filter(lambda t: t.is_running)
            .filter(worker.on_this_worker)
            .filter(worker.is_complete), )[0]


def auto_complete():
    for w in SUPPORTED_WORKERS:
        tasks = (read_complete_tasks_on_worker(w)
                 .subscribe(interface.mark_complete))


def auto_submit_root():
    (interface.read_all()
     .filter(lambda t: t.is_before_submit)
     .filter(lambda t: t.is_root)
     .subscribe(interface.mark_submit))


def auto_submit_chain():
    (interface.read_all()
     .filter(lambda t: t.is_pending)
     .flat_map(interface.dependencies)
     .filter(lambda t: t.is_before_submit)
     .subscribe(interface.mark_submit))


def is_dependencies_complete(task):
    return (interface.dependencies(task)
            .all(lambda t: t.is_complete))


def start(task):
    for w in SUPPORTED_WORKERS:
        if w.on_this_worker(task):
            w.run(task)


def auto_start():
    tasks = (interface.read_all()
             .filter(lambda t: t.is_pending))
    (rx.Observable.zip_array(tasks,
                             tasks.flat_map(is_dependencies_complete))
     .filter(lambda x: x[1])
     .map(lambda x: x[0])
     .subscribe(lambda t: start(t)))


# def cycle():
#     auto_complete()
#     auto_submit_root()
#     auto_submit_chain()
#     auto_start()
#     print('Cycle.')


# def launch_deamon(interval=1000):
#     configs = provider.get_or_create_service('config')
#     configs.set_config_by_name_key('database', 'use_web_api', True)
#     rx.Observable.interval(interval).start_with(0).subscribe(
#         on_next=lambda t: cycle(), on_error=lambda e: print(e))



from fs.osfs import OSFS
from .model import File


def touch(f):
    with OSFS('/') as fs:
        fs.create(f.path.rel)


def mv(f, path_new, overwrite=True):
    with OSFS('/') as fs:
        fs.move(f.path.rel, path_new, overwrite)


def cp(f, path_new, overwrite=True):
    with OSFS('/') as fs:
        fs.copy(f.path.rel, path_new, overwrite)


def rm(f):
    with OSFS('/') as fs:
        fs.remove(f.path.rel)



from .base import Detector2D
import numpy as np


class Detector2DParallelRing(Detector2D):
    def __init__(self, sensors=None, views=None, *, nb_sensors=None, sensor_width=None, nb_views=None):
        if sensors is None:
            sensors = np.arange(1, nb_sensors + 1) * sensor_width - \
                ((nb_sensors / 2 + 0.5) * sensor_width)
        if views is None:
            views = np.linspace(0, 2 * np.pi, nb_views)
        self._sensors = np.array(sensors)
        self._views = np.array(views)
        if self._sensors.ndim != 1:
            raise ValueError('Sensor sensors spec is required to be one dimension, got {}.'.format(
                self._sensors.ndim))
        if self._views.ndim != 1:
            raise ValueError('Sensor views spec is required to be one dimension, got {}.'.format(
                self._views.ndim))

    @property
    def nb_sensors(self):
        return len(self._sensors)

    @property
    def sensor_width(self):
        return np.mean(self._sensors[1:] - self._sensors[:-1])

    @property
    def sensors(self):
        return self._sensors

    @property
    def nb_views(self):
        return len(self._views)

    @property
    def views(self):
        return self._views

    def assert_fit(self, data):
        if self.nb_views != data.shape[0] or self.nb_sensors != data.shape[1]:
            msg = "Shape of sinogram {} is not consisted with detector: nb_sensors: {}, nb_views: {}."
            raise ValueError(msg.format(data.shape,
                                        self.nb_sensors, self.nb_views))



class DXPYTypeError(TypeError):
    def __init__(self, required_type, got_type_or_object, object_name="Something"):
        if not isinstance(got_type_or_object, type):
            got_type_or_object = type(got_type_or_object)
        super("{} is required to be instance of {}, got {}.".format(object_name,
                                                                    required_type,
                                                                    got_type_or_object))



import tensorflow as tf
from dxpy.configs import configurable, ConfigsView
# from ..config import get_configs_view, config
from dxpy.learn.config import config
from .base import NodeKeys

def mean_square_error(label, data):
    with tf.name_scope('mean_squared_error'):
        return tf.sqrt(tf.reduce_mean(tf.square(label - data)))


def poission_loss(label, data, *, compute_full_loss=False):
    with tf.name_scope('poission_loss'):
        label = tf.maximum(label, 0.0)
        # return log_possion_loss(tf.log(label), data)
        return tf.reduce_mean(tf.keras.losses.poisson(label, data))
                                


def log_possion_loss(log_label, data, *, compute_full_loss=False):
    """
    log_label: log value of expectation (inference)
    data: Poission sample
    """
    with tf.name_scope('log_poission_loss'):
        data = tf.maximum(data, 0.0)
        return tf.reduce_mean(tf.nn.log_poisson_loss(data, log_label, compute_full_loss))

from ..model.base import Model

class PoissionLossWithDenorm(Model):
    """
    Inputs:
    NodeKeys.INPUT: infer
    NodeKeys.LABEL: label
    Outputs:
    NodeKeys.OUTPUT: scalar loss
    """
    @configurable(config, with_name=True)
    def __init__(self, name, inputs, with_log=False, threshold=10, mean=0.0, std=1.0, **kw):
        super().__init__(name, inputs=inputs, with_log =with_log, threshold=threshold, mean=mean, std=std, **kw)
    def _kernel(self, feeds):
        from dxpy.debug.utils import dbgmsg
        dbgmsg(self.param('mean'))
        dbgmsg(self.param('std'))

        label = feeds[NodeKeys.LABEL]
        infer = feeds[NodeKeys.INPUT]
        with tf.name_scope('denorm_white'):
            label = label * self.param('std') + self.param('mean')
            infer = infer * self.param('std') + self.param('mean')
        if self.param('with_log'):
            with tf.name_scope('denorm_log_for_data'):
                infer = tf.exp(infer)
            with tf.name_scope('loss'):
                loss = log_possion_loss(label, infer)
        else:
            with tf.name_scope('loss'):
                loss = poission_loss(label, infer) 
        return {NodeKeys.MAIN: loss}


import unittest
from fs.memoryfs import MemoryFS


def make_test_fs(fs):
    fs.makedir('sub.0')
    with fs.opendir('sub.0') as sub:
        sub.makedir('subsub.0')
        with sub.opendir('subsub.0') as subsub:
            subsub.touch('test.txt')


class TestCaseWithMemoryFS(unittest.TestCase):
    def setUp(self):
        self.fs = MemoryFS()

    def tearDown(self):
        self.fs.close()



import tensorflow as tf


def random_crop(input_, target_shape, name='random_crop'):
    from dxpy.tensor.transform import random_crop_offset
    from ..tensor import ensure_tensor, shape_as_list
    with tf.name_scope(name):
        input_shape = shape_as_list(input_)
        target_shape = shape_as_list(target_shape)
        random_offset = tf.py_func(random_crop_offset,
                                   [input_shape, target_shape], tf.int64)
        return tf.slice(input_, random_offset, target_shape)


def align_crop(input_, target, offset=None, name='align_crop'):
    from ..tensor import shape_as_list
    with tf.name_scope(name):
        shape_input = shape_as_list(input_)
        shape_output = shape_as_list(target)
        if offset is None:
            offset = [0] + [(shape_input[i] - shape_output[i]) // 2
                            for i in range(1, 3)] + [0]
        return tf.slice(input_, offset, shape_output)


def boundary_crop(input_, offset=None, ratio=None, name="boundary_crop"):
    from ..tensor import shape_as_list
    with tf.name_scope(name):
        shape = shape_as_list(input_)
        if len(offset) == 2:
            offset = [0] + list(offset) + [0]
        shape_output = [s - 2 * o for s, o in zip(shape, offset)]
        return tf.slice(input_, offset, shape_output)





import json
from dxpy.time.utils import now, delta, strf, strp


class TimeStamps:
    """ A dict of time stamps
    """

    def __init__(self, time_stamps=None):
        if isinstance(time_stamps, TimeStamps):
            time_stamps = time_stamps.time_stamps
        if not isinstance(time_stamps, dict):
            raise TypeError(
                "time_stamps must be a dict, not {}.".format(type(time_stamps)))
        self.time_stamps = time_stamps


class Start(TimeStamps):
    def __init__(self, start=None):
        if start is None:
            start = now()
        super(Start, self).__init__({'start': start})

    @property
    def start(self):
        return self.time_stamps['start']


class Duration(Start):
    def __init__(self, start, end):
        super(__class__, self).__init__(start=start)
        self.time_stamps['end'] = end

    @property
    def end(self):
        return self.time_stamps['end']

    @property
    def length(self):
        return self.end - self.start

    def update_end(self, end):
        self.time_stamps['end'] = end


class Run(Start):
    def __init__(self, start, run=None):
        super(__class__, self).__init__(start=start)
        if run is None:
            run = delta()
        self.run = run

    def update_run(self):
        self.run = now() - self.start


class Progress(Duration, Run):
    def __init__(self, start, end, run=None):
        Duration.__init__(self, start=start, end=end)
        if run is None:
            run = delta()
        self.run = run

    @property
    def progress(self):
        return self.run / self.length


class TaskStamp(TimeStamps):
    def __init__(self, create=None, start=None, end=None):
        super(__class__, self).__init__({
            'create': create,
            'start': start,
            'end': end
        })

    @property
    def create(self):
        return self.time_stamps['create']

    @property
    def start(self):
        return self.time_stamps['start']

    @property
    def end(self):
        return self.time_stamps['end']

    def to_json(self):
        dct = {
            'create': strf(self.create),
            'start': strf(self.start),
            'end': strf(self.end)
        }
        return json.dumps(dct)

    @classmethod
    def from_json(cls, s):
        dct = json.loads(s)
        return cls(create=strp(dct['create']), start=dct['start'], end=dct['end'])

    @classmethod
    def create_now(cls):
        return cls(create=now(), start=None, end=None)



from pathlib import Path
from flask import Flask
from flask_sqlalchemy import SQLAlchemy

def get_or_create_db(cls, db_path, db_file, name=__name__):
    if not db_file.endswith('.db'):
        db_file += '.db'
    full_path = Path(db_path) / db_file    
    app = Flask(name)
    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:////'+str(full_path.absolute())
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    db = SQLAlchemy(app)
    if not full_path.exists():    
        cls.get_db_model(db)
        db.create_all()
    return db




"""
Decorate functions foo(*args, **kw) to provide `kw` from config objects.
Useage:
Case 1:
@configurable(c)
def foo(xx=None, yy=None, zz=None, name='default_name'):
    pass
Will try to load config by c
Case 2:
@configurable(c)
def foo(xx=None, yy=None, zz=None, name='default_name'):
    pass
Will try to load config by c['defautl_name']

Case 3:
@configurable(c.get('conf_name'))
def foo(**kw):
    pass
Will try to load config by c['conf_name']
"""
import inspect
from functools import wraps

KEY_NAME = 'name'


def parse_configs(_func, *args, _config_object, **kw):
    sig = inspect.signature(_func)
    parameter_keys_not_in_args = list(sig.parameters.keys())[len(args):]
    kw_refined = dict()
    for k in parameter_keys_not_in_args:
        if k in kw and kw[k] is not None:
            kw_refined[k] = kw[k]
            continue
        if _config_object[k] is not None:
            kw_refined[k] = _config_object[k]
            continue
    left_kw = dict()
    for k in kw:
        if kw[k] is not None and not k in kw_refined:
            left_kw[k] = kw[k]
    # for v in sig.parameters.values():
        # if v.kind == inspect.Parameter.VAR_KEYWORD:
            # kw_refined[v.name] = left_kw
    ba = sig.bind(*args, **kw_refined, **left_kw)
    ba.apply_defaults()
    args = dict(ba.arguments)
    args.update(left_kw)
    return ba


def get_name(_func, *args, **kw):
    sig = inspect.signature(_func)
    ba = sig.bind_partial(*args, **kw)
    ba.apply_defaults()
    return ba.arguments[KEY_NAME]

class configurable:
    def __init__(self, configs_object=None, with_name=False):
        from ._viewer import ConfigsView
        if configs_object is None:
            configs_object = dict()
        if isinstance(configs_object, dict):
            configs_object = ConfigsView(configs_object)
        self._c = configs_object
        self._with_name = with_name

    def __call__(self, _func):
        @wraps(_func)
        def wrapper(*args, **kw):
            config = self._c
            if self._with_name:
                config = config.get(get_name(_func, *args, **kw))
            ba = parse_configs(_func, *args, **kw, _config_object=config)
            return _func(*ba.args, **ba.kwargs)
        return wrapper



import tensorflow as tf
from dxpy.configs import configurable
from dxpy.learn.config import config



@configurable(config, with_name=True)
def start_dataset_server(cluster_spec_file='cluster.yml', job_name='dataset', task_index=0, dataset_name=None, name='cluster/dataset/task0'):
    """
    Args:
        func: functions which returns dataset
    """
    from .cluster import get_cluster_spec, get_server
    cluster = get_cluster_spec(cluster_spec_file)
    server = tf.train.Server(cluster, job_name=job_name, task_index=task_index)
    get_dist_dataset(name=name)
    server.join()

@configurable(config, with_name=True)
def get_dist_dataset(job_name='dataset', task_index=0, dataset_config=None, name='cluster/dataset/task0'):
    from dxpy.learn.dataset import get_dataset
    with tf.device('/job:{}/task:{}'.format(job_name, task_index)):
        dataset = get_dataset(name=dataset_config)
    return dataset



import tensorflow as tf
from ..graph import Graph, NodeKeys
from contextlib import contextmanager
from dxpy.configs import configurable
from dxpy.learn.config import config

_session = None


def set_default_session(session):
    global _session
    _session = session


def get_default_session():
    if _session is None:
        set_default_session(tf.get_default_session())
    return _session


class Session(Graph):
    def __init__(self, name='session', **config):
        super().__init__(name, **config)
        self._create_session()

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'allow_growth': True,
            'log_device_placement': False
        }, super()._default_config())

    def _create_session(self):
        config = tf.ConfigProto()
        if self.param('allow_growth'):
            config.gpu_options.allow_growth = True
        if self.param('log_device_placement'):
            config.log_device_placement = True
        self.register_main_node(tf.Session(config=config))

    def run(self, tensors, feed_dict=None):
        with self.nodes[NodeKeys.MAIN].as_default():
            return self.nodes[NodeKeys.MAIN].run(tensors, feed_dict)

    def run_func(self, func):
        with self.nodes[NodeKeys.MAIN].as_default():
            return func()

    @property
    def session(self):
        return self.nodes[NodeKeys.MAIN]

    @property
    def graph(self):
        return self.session.graph

    @contextmanager
    def as_default(self):
        with self.nodes[NodeKeys.MAIN].as_default():
            yield self.nodes[NodeKeys.MAIN]

    def post_session_created(self):
        self.as_tensor().run(tf.global_variables_initializer())


class SessionDist(Session):
    def __init__(self, name='session', target=None, **kw):
        self._target = target
        super().__init__(name=name, **kw)

    def _create_session(self):
        config = tf.ConfigProto()
        if self.param('allow_growth'):
            config.gpu_options.allow_growth = True
        if self.param('log_device_placement'):
            config.log_device_placement = True

        self.register_main_node(tf.Session(self._target, config=config))






""" Task APIs
"""

from ..model.task import Task
from ..interface import read, read_all


class SlurmCreator:
    @classmethod
    def create(cls, workdir, filename):
        from ..interface import create
        from ..model.creators import task_slurm
        return create(task_slurm(workdir, filename))

    @classmethod
    def create_group(cls, workdirs, filenames, depens):
        from ..interface import create, create_graph
        from ..model.creators import task_graph
        tasks = [create(w, f, no_create=True)
                 for w, f in zip(workdirs, filenames)]
        g = create_graph(tasks, depens)
        return create_graph(g)


def submit(tid):
    from ..interface import mark_submit
    mark_submit(read(tid))


def delete(tid):
    from ..interface import delete
    delete(tid)



#import ptvsd
#ptvsd.enable_attach("ptvsd", address = ('0.0.0.0', 3000))

# Enable the line of source code below only if you want the application to wait until the debugger has attached to it
#ptvsd.wait_for_attach()

import tensorflow as tf
import numpy as np
from dxpy.learn.utils.debug import dummy_image, write_graph
from dxpy.learn.model.cnn.super_resolution import SuperResolutionBlock, SRKeys
from dxpy.learn.graph import NodeKeys
with tf.device('/gpu:1'):
    image = dummy_image()
    label = dummy_image(shape=[14, 14])
    reps = dummy_image(shape=[14, 14], nb_channel=32)
    m = SuperResolutionBlock('srb', {NodeKeys.INPUT: image, NodeKeys.LABEL: label})
    result = m()
    write_graph()


from ..graph import Graph


class GraphFromDict(Graph):
    def __init__(self, nodes, *, name='dict_graph', **config):
        super().__init__(name, **config)
        for n, v in nodes.items():
            self.register_node(n, v)


class GraphProxy(Graph):
    def __init__(self, origin, proxy_map, *, name='proxy', **config):
        """
        Args:
            proxy_map:  dict of new_name: old_name
        """
        super().__init__(name, **config)
        for n, v in proxy_map:
            self.register_node(n, origin[v])



def path_of_script_folder_of_module(module, relative_path='../scripts'):
    from dxpy.filesystem import Path
    return Path(Path(module.__file__) / relative_path)


def path_of_global_defaults(root_module='dxpy'):
    from dxpy.filesystem import Path
    return Path(Path('~') / '.{m}'.format(m=root_module))



import rx
import jinja2
from dxpy.filesystem import Path


class Service:
    """ Pure logic module
    """
    pass


def template_dir():
    return Path(Path(__file__).father)


def cli_template(name):
    # TODO: Add jinja2 implementation
    with open((template_dir() / 'cli' / 'base.py').abs) as fin:
        tpl = jinja2.Template(fin.read())
    return tpl.render(name=name)


def web_template(name):
    class_names = []
    (rx.Observable.from_(name.split('_'))
     .map(lambda s: s.capitalize())
     .subscribe(class_names.append))
    class_name = ''.join(class_names)
    resource_name = class_name + 'Resource'
    resources_name = class_name + 'sResource'
    imports = """#Auto generated web.py
from flask import Flask, make_response, request, Response
from flask_restful import Resource, reqparse, Api
from . import api
"""

    resource = """
class {name}(Resource):
    def get(self, id):
        try:
            return Response(api.read(id), 200, mimetype="application/json")
        except Exception as e:
            return str(e), 404

    def put(self, id):
        try:
            obj = request.form['obj']
            return Response(api.update(task), 201, mimetype="application/json")
        except Exception as e:
            return str(e), 404

    def delete(self, id):
        try:
            return Response(api.delete(id), 200, mimetype="application/json")
        except Exception as e:
            return str(e), 404
""".format(name=resource_name)

    resources = """
class {name}(Resource):
    import json

    def get(self):
        try:
            objs = []
            api.get_all().subscribe(lambda o: objs.append(o))
            return Response(__class__.json.dumps(objs), 200, mimetype="application/json")
        except Exception as e:
            return str(e), 404

    def post(self):
        obj = request.form['obj']
        result = api.create(obj)
        return Response(__class__.json.dumps({result}), 201, mimetype="application/json")
""".format(name=resources_name, result=r"{'result': result}")

    add_api = """
def add_api(api, root):
    from .config import c
    api.add_resource({name}, c.get_url_endpoint(root))
    api.add_resource({names}, c.get_url_endpoint(root, plural=True))
""".format(name=resource_name, names=resources_name)

    launch_server = """
def launch_server():
    from .config import c
    app = Flask(__name__)
    api = Api(app)
    add_api(api, c.get_root())
    app.run(host=c.ip, port=c.port, debug=c.debug)
"""

    return ''.join([imports, resource, resources, add_api, launch_server])


class SlurmScript:
    def __init__(self, name, path):
        self.path = path
        self.name = name

    def make(self):
        from fs.osfs import OSFS
        with OSFS(self.path) as fs:
            if fs.exists(self.name):
                d = fs.opendir(self.name)
            else:
                d = fs.makedir(self.name)
            with d.open(self.name, 'w') as fout:
                
            d.close()


class Component:
    """ Moduel include api, cli and web interface and an service for logics.
    """

    def __init__(self, name, path):
        self.path = path
        self.name = name

    def _make_init(self, fs):
        fs.touch('__init__.py')

    def _make_api(self, fs):
        with fs.open('api.py', 'w') as fout:
            print('from . import service', file=fout)

    def _make_cli(self, fs):
        with fs.open('cli.py', 'w') as fout:
            print(cli_template(self.name), file=fout)

    def _make_web(self, fs):
        with fs.open('web.py', 'w') as fout:
            print(web_template(self.name), file=fout)

    def _make_service(self, fs):
        with fs.open('service.py', 'w') as fout:
            pass

    def _make_config(self, fs):
        with fs.open('config.py', 'w') as fout:
            pass

    def _make_exceptions(self, fs):
        with fs.open('exceptions.py', 'w') as fout:
            pass

    def _make_tests(self, fs):
        with fs.open('test_{name}.py'.format(name=self.name), 'w') as fout:
            pass

    def make(self):
        from fs.osfs import OSFS
        with OSFS(self.path) as fs:
            if fs.exists(self.name):
                d = fs.opendir(self.name)
            else:
                d = fs.makedir(self.name)
            self._make_init(d)
            self._make_api(d)
            self._make_cli(d)
            self._make_web(d)
            self._make_config(d)
            self._make_service(d)
            self._make_exceptions(d)
            self._make_tests(d)
            d.close()



import tensorflow as tf
from ..base import Model, NodeKeys
from .blocks import StackedConv2D
from .residual import StackedResidual
from ..image import align_crop
from dxpy.configs import configurable
from ...config import config


class SRKeys:
    REPRESENTS = 'reps'
    RESIDUAL = 'resi'
    ALIGNED_LABEL = 'aligned_label'
    INTERP = 'interp'
    POI_LOSS = 'poi_loss'
    MSE_LOSS = 'mse_loss'


class BuildingBlocks:
    STACKEDCONV = 'stacked_conv'
    RESCONV = 'res_conv'
    RESINCEPT = 'res_incept'
    INTERP = 'interp'


class SuperResolution2x(Model):
    def __init__(self, name, inputs, *,
                 building_block: str=None,
                 nb_layers: int=None,
                 filters: int=None,
                 boundary_crop=None,
                 **config):
        super().__init__(name, inputs,
                         building_block=building_block,
                         nb_layers=nb_layers,
                         filters=filters,
                         boundary_crop=boundary_crop,
                         **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'building_block': BuildingBlocks.STACKEDCONV,
            'nb_layers': 10,
            'filters': 32,
            'boundary_crop': [4, 4]
        }, super()._default_config())

    def _kernel(self, feeds):
        from ..image import resize, boundary_crop, mean_square_error
        with tf.variable_scope('input'):
            u = resize(feeds[NodeKeys.INPUT], (2, 2))
            if SRKeys.REPRESENTS in feeds:
                r = resize(feeds[SRKeys.REPRESENTS], (2, 2))
                r = align_crop(r, u)
                r = tf.concat([r, u], axis=3)
            else:
                r = tf.layers.conv2d(u, self.param('filters'), 5, name='stem')
        if self.param('building_block') == BuildingBlocks.STACKEDCONV:
            x = StackedConv2D('kernel', r,
                              nb_layers=self.param('nb_layers'),
                              filters=self.param('filters'))()
        elif self.param('building_block') == BuildingBlocks.RESCONV:
            x = StackedResidual('kernel', r, self.param(
                'nb_layers'), StackedResidual.STACKED_CONV_TYPE)()
        elif self.param('building_block') == BuildingBlocks.RESINCEPT:
            x = StackedResidual('kernel', r, nb_layers=self.param('nb_layers'),
                                block_type=StackedResidual.INCEPT_TYPE)()
        with tf.variable_scope('inference'):
            res = tf.layers.conv2d(x, 1, 3, padding='same')
            res = boundary_crop(res, self.param('boundary_crop'))
            u_c = align_crop(u, res)
            y = res + u_c
        result = {NodeKeys.INFERENCE: y,
                  SRKeys.REPRESENTS: x,
                  SRKeys.RESIDUAL: res,
                  SRKeys.INTERP: u_c}
        if NodeKeys.LABEL in feeds:
            with tf.name_scope('loss'):
                aligned_label = align_crop(feeds[NodeKeys.LABEL], y)
                l = mean_square_error(aligned_label, y)
            result.update({NodeKeys.LOSS: l,
                           SRKeys.ALIGNED_LABEL: aligned_label})
        return result


from ...config import config as cfg


class SuperResolutionBlock(Model):
    """
    Inputs:
        NodeKeys.INPUT: input low resolution image
        NodeKeys.LABEL: [Optional] high resolution label image
        SRKeys.REPRESENTS: [Optional] low resolution feature representations
    Outputs:
        NodeKeys.INFERENCE: high resolution image (cropped)
        NodeKeys.LOSS: REQUIRES: Inputs[NodeKeys.LABEL], loss of inference
        SRKeys.REPRESENTS: represents befefore inference image
    Sub-models:
        'kernel': One of StackedConv2D or StackedResidual 
    """

    @configurable(cfg, with_name=True)
    def __init__(self, name, inputs, *,
                 building_block: str=BuildingBlocks.STACKEDCONV,
                 filters: int=32,
                 boundary_crop=(4, 4),
                 down_sample_ratio=(2, 2),
                 boundary_crop_ratio=0.1,
                 with_poi_loss=False,
                 poi_loss_weight=0.0,
                 mse_loss_weight=1.0,
                 **kw):
        super().__init__(name, inputs,
                         building_block=building_block,
                         filters=filters,
                         down_sample_ratio=down_sample_ratio,
                         boundary_crop=boundary_crop,
                         boundary_crop_ratio=boundary_crop_ratio,
                         with_poi_loss=with_poi_loss,
                         poi_loss_weight=poi_loss_weight,
                         mse_loss_weight=mse_loss_weight,
                         **kw)

    def _crop_size(self, input_=None):
        from ..tensor import shape_as_list
        if self.param('boundary_crop') == 'relative':
            shape = shape_as_list(input_)
            ratio = self.param('boundary_crop_ratio')
            return [int(shape[0] * ratio), int(shape[1] * ratio)]
        else:
            return self.param('boundary_crop')

    def _input(self, feeds):
        """
        """
        # TODO： add check for input shape
        from ..image import resize
        from ..tensor import shape_as_list
        with tf.variable_scope('input'):
            u = resize(feeds[NodeKeys.INPUT], self.param('down_sample_ratio'))
            if self.param('building_block') == BuildingBlocks.INTERP:
                return u, None
            if SRKeys.REPRESENTS in feeds:
                r = resize(feeds[SRKeys.REPRESENTS],
                           self.param('down_sample_ratio'))
                r = align_crop(r, u)
                r = tf.concat([r, u], axis=3)
            else:
                r = tf.layers.conv2d(u, self.param(
                    'filters'), 5, name='stem', padding='same')
            if NodeKeys.LABEL in feeds:
                l = feeds[NodeKeys.LABEL]
            else:
                l = None
            return u, r, l

    def _inference(self, represents, upsampled):
        from ..image import boundary_crop, align_crop

        with tf.variable_scope('inference'):
            upsampled = boundary_crop(upsampled, self._crop_size(upsampled))
            if self.param('building_block') == BuildingBlocks.INTERP:
                return {NodeKeys.INFERENCE: upsampled}
            residual = tf.layers.conv2d(represents, 1, 3, padding='same')
            residual = align_crop(residual, upsampled)
            inference = residual + upsampled
        return {NodeKeys.INFERENCE: inference,
                SRKeys.RESIDUAL: residual,
                SRKeys.INTERP: upsampled}

    def _kernel_stacked_convolution(self, represents):
        return StackedConv2D(self.name / 'kernel', represents)()

    def _kernel_stacked_residual(self, represents):
        from ..cnn.residual import StackedResidualv2
        if self.param('building_block') == BuildingBlocks.RESCONV:
            block_type = StackedResidual.STACKED_CONV_TYPE
        else:
            block_type = StackedResidual.INCEPT_TYPE
        return StackedResidualv2(self.name / 'kernel', represents, block_type=block_type)()

    def _mse_loss(self, label, infer):
        from ..image import mean_square_error
        return mean_square_error(label, infer)

    def _poi_loss(self, label, infer):
        from ..losses import PoissionLossWithDenorm
        return PoissionLossWithDenorm(self.name / 'loss' / 'poi',
                                      {NodeKeys.INPUT: infer,
                                       NodeKeys.LABEL: label}).as_tensor()

    def _loss(self, label, infer):
        from ..image import mean_square_error, align_crop
        from ..losses import PoissionLossWithDenorm
        if label is not None:
            with tf.name_scope('loss'):
                aligned_label = align_crop(label, infer)
                loss_mse = self._mse_loss(
                    aligned_label, infer) * self.param('mse_loss_weight')
                loss = loss_mse
                if self.param('with_poi_loss'):
                    loss_poi = self._poi_loss(
                        aligned_label, infer) * self.param('poi_loss_weight')
                    loss = loss + loss_poi
                result = {NodeKeys.LOSS: loss,
                          SRKeys.MSE_LOSS: loss_mse,
                          SRKeys.ALIGNED_LABEL: aligned_label}
                if self.param('with_poi_loss'):
                    result.update({SRKeys.POI_LOSS: loss_poi})
                return result
        else:
            return dict()

    def _kernel(self, feeds):
        upsampled, represents, label = self._input(feeds)
        if self.param('building_block') == BuildingBlocks.INTERP:
            x = upsampled
        elif self.param('building_block') == BuildingBlocks.STACKEDCONV:
            x = self._kernel_stacked_convolution(represents)
        elif self.param('building_block') in [BuildingBlocks.RESCONV, BuildingBlocks.RESINCEPT]:
            x = self._kernel_stacked_residual(represents)
        result = {SRKeys.REPRESENTS: x}
        result.update(self._inference(x, upsampled))
        result.update(self._loss(label, result[NodeKeys.INFERENCE]))
        return result


class SuperResolutionMultiScale(Model):
    """
    Inputs:
        'input/image1x'
        NodeKeys.LABEL: [Optional] high resolution label image
        SRKeys.REPRESENTS: [Optional] low resolution feature representations
    Outputs:
        NodeKeys.INFERENCE: high resolution image (cropped)
        NodeKeys.LOSS: <REQUIRED: Inputs[NodeKeys.LABEL], loss of inference
        SRKeys.REPRESENTS: represents befefore inference image
    """

    def __init__(self, name, inputs, nb_down_sample, *, share_model=None, **config):
        super().__init__(name, inputs, nb_down_sample=nb_down_sample,
                         share_model=share_model, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'nb_down_sample': 3,
            'loss_weight': 1.0,
            'share_model': False
        }, super()._default_config())

    @classmethod
    def multi_scale_input(cls, images, nb_down_sample=None, labels=None):
        if nb_down_sample is None:
            nb_down_sample = len(images) - 1
        if labels is None:
            labels = images
        inputs = {}
        for i in range(nb_down_sample + 1):
            inputs.update({'input/image{}x'.format(2**i): images[i]})
        for i in range(nb_down_sample):
            inputs.update({'label/image{}x'.format(2**i): labels[i]})
        return inputs

    def _get_node(self, node_type, down_sample_ratio, source=None):
        if source is None:
            source = self.nodes
        name = "{}/image{}x".format(node_type, 2**down_sample_ratio)
        return source[name]

    def _kernel(self, feeds):
        
        x = None
        r = None
 
        losses = []
        for i_down_sample in reversed(range(1, self.param('nb_down_sample') + 1)):
            if x is None:
                x = self._get_node('input', i_down_sample, feeds)

            mid_result = SuperResolution2x('sr2x_{}'.format(i_down_sample),
                                           {NodeKeys.INPUT: x,
                                            NodeKeys.LABEL: self._get_node('label', i_down_sample - 1, feeds),
                                            SRKeys.REPRESENTS: r})()
            x = mid_result[NodeKeys.INFERENCE]
            if NodeKeys.LOSS in mid_result:
                losses.append(mid_result[NodeKeys.LOSS])
            r = mid_result[SRKeys.REPRESENTS]
        result = {k: mid_result[k] for k in mid_result}
        if len(losses) > 0:
            with tf.name_scope('loss'):
                if not isinstance(self.param('loss_weight'), (list, tuple)):
                    lw = [self.param('loss_weight')] * \
                        self.param('nb_down_sample')
                else:
                    lw = self.param('loss_weight')
                for i, l in enumerate(losses):
                    losses[i] = losses[i] * lw[i]
                result[NodeKeys.LOSS] = tf.add_n(losses)
        return result


import typing


class SuperResolutionMultiScalev2(Model):
    """
    Inputs:
    NodeKeys.Input: Low resolution inputs
    'input/image[1, 2, 4, 8]x': input images, only the one with largest down sample ratio is required,
    others will be ignored. 
    'input/clean[1, 2, 4, 8]x': labels, may be noisy.
    Outputs:
    NodeKeys.INFERENCE: inference image
    'interp': interp image(cropped)
    'aligned_label': label image (cropped)
    'mse_loss': mse loss (if with_poi_loss)
    'poi_loss': poi loss (if with_poi_loss)
    NodeKeys.LOSS: scalar loss 
    SRKeys.ALIGNED_LABEL: cropped label (useful for calculating residual)
    """
    @configurable(config, with_name=True)
    def __init__(self, name, inputs, *,
                 nb_down_sample: int=3,
                 share_model=False,
                 loss_weight: typing.TypeVar('T', float, typing.List[float]) = 2.0,
                 **kw):
        super().__init__(name, inputs,
                         nb_down_sample=nb_down_sample,
                         share_model=share_model,
                         loss_weight=loss_weight,
                         **kw)
    @classmethod
    def multi_scale_input(cls, images, nb_down_sample=None, labels=None):
        """
        Helper function to make valid images from varient input schemes.
        Args:
        -   images: list of input tensors. `images[-1]` should usually be the input low resolution image.
            For those case considering noise, images maybe used as training labels
        -   nb_down_sample: number of down sample operations, if is `None`, `len(images) -1` will be used.
        -   labels: list of label tensors. 
        """
        
        if nb_down_sample is None:
            nb_down_sample = len(images) - 1
        if labels is None:
            labels = images
        inputs = {}
        for i in range(nb_down_sample + 1):
            inputs.update({'input/image{}x'.format(2**i): images[i]})
        for i in range(nb_down_sample):
            inputs.update({'label/image{}x'.format(2**i): labels[i]})
        return inputs

    def _get_node(self, node_type, down_sample_level, source=None):
        """
        Get nodes from source dict-like object.
        Args:
        -   node_type: 'image' or 'label',
        -   down_sample_level: required down sample level, 
        -   source: dict-like object, if is `None`, use `self.nodes`
        """

        if source is None:
            source = self.nodes
        name = "{}/image{}x".format(node_type, 2**down_sample_level)
        return source[name]

    def _kernel(self, feeds):
        from ..image import resize
        x = None
        r = None
        itp = None
        losses = []
        losses_mse = []
        losses_poi = []
        for i_down_sample in reversed(range(1, self.param('nb_down_sample') + 1)):
            if x is None:
                x = self._get_node('input', i_down_sample, feeds)
            if itp is None:
                itp = x
            srb = SuperResolutionBlock(self.name / 'srb_{}'.format(i_down_sample),
                                              {NodeKeys.INPUT: x,
                                               NodeKeys.LABEL: self._get_node('label', i_down_sample - 1, feeds),
                                               SRKeys.REPRESENTS: r})
            mid_result = srb()
            x = mid_result[NodeKeys.INFERENCE]
            itp = resize(itp, srb.param('down_sample_ratio'))
            itp = align_crop(itp, x)
            if NodeKeys.LOSS in mid_result:
                losses.append(mid_result[NodeKeys.LOSS])
                if SRKeys.POI_LOSS in mid_result:
                    losses_poi.append(mid_result[SRKeys.POI_LOSS])
                    losses_mse.append(mid_result[SRKeys.MSE_LOSS])
            r = mid_result[SRKeys.REPRESENTS]
        result = {k: mid_result[k] for k in mid_result}
        result[SRKeys.INTERP] = itp
        if len(losses) > 0:
            with tf.name_scope('loss'):
                if not isinstance(self.param('loss_weight'), (list, tuple)):
                    lw = [self.param('loss_weight')] * \
                        self.param('nb_down_sample')
                else:
                    lw = self.param('loss_weight')
                for i, l in enumerate(losses):
                    losses[i] = losses[i] * lw[i]
                    if len(losses_poi) > 0: 
                        losses_poi[i] = losses_poi[i] * lw[i]
                        losses_mse[i] = losses_mse[i] * lw[i]
                result[NodeKeys.LOSS] = tf.add_n(losses)
                if len(losses_poi) > 0:
                    result[SRKeys.POI_LOSS] = tf.add_n(losses_poi)
                    result[SRKeys.MSE_LOSS] = tf.add_n(losses_mse)
        return result



from dxpy.configs import configurable
from dxpy.learn.config import config
import tensorflow as tf
import yaml


def get_cluster_spec(config_filename, job_name=None):
    return tf.train.ClusterSpec(get_cluster_spec_raw(config_filename, job_name))

def get_cluster_spec_raw(config_filename, job_name=None):
    with open(config_filename) as fin:
        spec = yaml.load(fin)
    result = dict()
    for job_name in spec:
        result[job_name] = ["{}:{}".format(v['ip'], v['port']) for v in spec[job_name]]
    if job_name == 'dataset':
        valid_keys = ['dataset']
    elif job_name == 'ps':
        valid_keys = ['ps', 'dataset']
    elif job_name == 'worker':
        valid_keys = ['ps', 'dataset', 'worker']
    elif job_name in ['summary', 'saver']:
        valid_keys = ['ps', 'dataset', job_name]
    else:
        valid_keys = list(result.keys)
    # result = {k: result[k] for k in valid_keys}
    return result

def get_nb_tasks(cluster_filename, job_name):
    cfg = get_cluster_spec_raw(cluster_filename)
    return len(cfg[job_name])

def get_server(cluster_spec, job_name, task_index=0, config=None):
    if isinstance(cluster_spec, dict):
        cluster_spec = tf.train.ClusterSpec(cluster_spec)
    cluster = tf.train.ClusterSpec(cluster_spec)
    server = tf.train.Server(cluster, job_name, task_index, config=config)
    return server





import click


@click.group()
def run():
    pass


@click.command()
def start():
    """ start task database web service """
    from .deamon import DeamonService
    DeamonService.start()
    # input("Press any key to exit\n")


run.add_command(start)



import tensorflow as tf
from ..model import Model
from ..net import Net
from ..graph import Graph, NodeKeys


class MNISTSimpleConvModel(Model):
    def __init__(self, name='/model', image=None, label=None, **config):
        inputs = {'image': image, 'label': label}
        super(__class__, self).__init__(name, inputs, **config)

    @classmethod
    def _default_inputs(self):
        return {
            'image': {'shape': [None, 28, 28, 1]},
            'label': {'shape': [None, 10]}}

    def _kernel(self, feeds):
        x = feeds['image']
        result = dict()
        with tf.variable_scope('kernel'):
            x = tf.layers.conv2d(x, 32, 5, padding='same',
                                 name='conv1', activation=tf.nn.relu)
            x = tf.layers.conv2d(x, 32, 1, padding='same',
                                 name='pool1', strides=(2, 2), activation=tf.nn.relu)
            x = tf.layers.conv2d(x, 64, 3, padding='same',
                                 name='conv2', activation=tf.nn.relu)
            x = tf.layers.conv2d(x, 64, 1, padding='same',
                                 name='pool2', strides=(2, 2), activation=tf.nn.relu)
            with tf.name_scope('flatten'):
                x = tf.reshape(x, [-1, 7 * 7 * 64])
            x = tf.layers.dense(x, 1024, activation=tf.nn.relu, name='fc1')
            result['logits'] = tf.layers.dense(x, 10, name='fc2')
        with tf.name_scope('prediction'):
            result['prediction'] = tf.argmax(result['logits'], 1)
        if 'label' in feeds:
            with tf.name_scope('loss'):
                result['cross_entropy'] = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(
                    labels=feeds['label'], logits=result['logits']))
            with tf.name_scope('accuracy'):
                correct_prediction = tf.equal(
                    result['prediction'], tf.argmax(feeds['label'], 1))
                result['accuracy'] = tf.reduce_mean(
                    tf.cast(correct_prediction, tf.float32))
        return result


class MNISTSimpleNet(Net):
    def __init__(self, name='/net', image=None, label=None, **config):
        inputs = {'image': image, 'label': label}
        super(__class__, self).__init__(name, inputs, **config)

    @classmethod
    def _default_inputs(self):
        return {
            'image': {'shape': [None, 28, 28, 1]},
            'label': {'shape': [None, 10]}}

    def _tensors_need_summary(self):
        return {
            'loss': {
                'type': 'scalar',
                'tensor': self.tensor(NodeKeys.LOSS)
            },
            'accuracy': {
                'type': 'scalar',
                'tensor': self.tensor(NodeKeys.EVALUATE)
            },
            'image': {
                'type': 'image',
                'tensor': self.tensor('image')
            }
        }

    def _pre_kernel_post_inputs(self):

        self._main_model = MNISTSimpleConvModel(self.name / 'model',
                                                self.nodes['image'],
                                                self.nodes['label'], lazy_create=True)

    def _post_kernel_post_outputs(self):
        self.register_node(
            NodeKeys.LOSS, self._main_model['cross_entropy'])
        self.register_node(NodeKeys.INFERENCE,
                           self._main_model['prediction'])
        self.register_node(NodeKeys.EVALUATE,
                           self._main_model['accuracy'])
        super()._post_kernel_post_outputs()

    def _kernel(self, feeds=None):
        return self._main_model({'image': self.tensor('image'), 'label': self.tensor('label')})


def main():
    from dxpy.learn.dataset.mnist import MNISTLoadAll
    from dxpy.learn.train.summary_writer import SummaryWriter
    from dxpy.learn.scalar import create_global_scalars
    create_global_scalars()
    create_global_scalars()
    dataset = MNISTLoadAll()
    net = MNISTSimpleNet(
        '/net', image=dataset['image'], label=dataset['label'])
    summary = SummaryWriter(
        'train', net._tensors_need_summary(), path='./summary/train/')
    sess = tf.Session(config=tf.ConfigProto(log_device_placement=False))
    with sess.as_default():
        sess.run(tf.global_variables_initializer())
        dataset.post_session_created()
        net.post_session_created()
        summary.post_session_created()
        feeds = sess.run(
            {'image': dataset['image'], 'label': dataset['label']})
        print('initial accuracy', net.evaluate())
        for i in tqdm_notebook(range(10)):
            for _ in tqdm_notebook(range(100)):
                net.train()
            feeds = dataset.get_feed_dict()
            print('STEP={}'.format((i + 1) * 100), net.evaluate())
            summary.summary()
            net.save()
        summary.flush()



import os
import fs.path as fp


class Path:
    """
    Properties returns simple and basic types: str or bool
    """

    def __init__(self, path, ptype=None):
        """
        Args:
            path: any object which is convertable to Path, including:
                str: raw path, url path, 
        """

        self._path = fp.normpath(str(path))
        if self._path.startswith('~'):
            self._path = os.environ['HOME'] + self._path[1:]

    def __str__(self):
        return self._path

    @property
    def raw(self):
        return self._path

    @property
    def abs(self):
        return fp.abspath(self._path)

    @property
    def rel(self):
        return fp.relpath(self._path)

    @property
    def is_abs(self):
        return fp.isabs(self._path)

    @property
    def basename(self):
        return fp.basename(self._path)

    @property
    def suffix(self):
        return fp.splitext(self._path)[1]

    def parts(self):
        result = tuple(fp.iteratepath(self._path))
        slash = fp.abspath('x')[:-1]
        if self.is_abs and result[0] != slash:
            result = tuple([slash] + list(result))
        return result

    def parent(self):
        return Path(fp.dirname(self._path))

    def __truediv__(self, name):
        return Path(fp.combine(self._path, name))

    def __add__(self, suffix):
        return Path(self._path + suffix)

    def __eq__(self, path):
        return self._path == str(Path(path))

    def __hash__(self):
        return hash(self._path)



class DXPYTypeError(TypeError):
    def __init__(self, required_type, got_object_or_type, prefix=None):
        from ._message import type_error_message
        if not isinstance(got_object_or_type, type):
            got_object_or_type = type(got_object_or_type)
        super().__init__(type_error_message(required_type, got_object_or_type, prefix))



from ruamel.yaml import YAML
from dxpy.filesystem import Path
yaml = YAML()


class ModuleConfigs:
    def __init__(self, module_name=None, path_user=None, path_workspace=None):
        self._load(module_name, path_user, path_workspace)

    def get_config(self):
        return self.data.get(['__WORKSPACE__'])

    def reload(self, path_workspace):
        self._load(self.module_name, self.path_user, path_workspace)

    def _load(self, module_name, path_user, path_workspace):
        from dxpy.collections import TreeDict
        self.module_name = module_name
        path_user = self._refine_path_user(path_user)
        self.path_user = self._path_config_file(path_user)
        self.path_workspace = self._path_config_file(path_workspace)
        self.data = TreeDict(self._load_config_file(self.path_user))
        self.data['__WORKSPACE__'] = TreeDict(
            self._load_config_file(self.path_workspace))

    def _refine_path_user(self, path_user):
        import os
        if isinstance(path_user, str) and not path_user.startswith('/'):
            return Path(os.environ['HOME']) / '.dxl' / path_user
        else:
            return Path(path_user)

    def _path_config_file(self, path):
        from dxpy.filesystem import Directory, File
        if path is None:
            return None
        filename = Path(path)
        if self.module_name is not None and Directory(filename).exists:
            possible_suffix = ['', '.yml', '.yaml', '.json']
            for sfx in possible_suffix:
                name = '{0}{1}'.format(self.module_name, sfx)
                if File(filename / name).exists:
                    filename = filename / name
                    break
        if not File(filename).exists:
            return None
        return Path(filename).abs

    def _load_config_file(self, path):
        if path is None:
            return dict()
        with open(path) as fin:
            dct = yaml.load(fin)
            if dct is None:
                dct = dict()
            result = dict()
            for k in dct:
                result[k] = dct[k]
            return result



"""
Module of *raw* datasets, thus only handles reading dataset files (usually .h5 or .tfrecord),
parse records/examples into tensors in dict-like collection, no normalization in principle.

However some properties may be provided, like mean or standard deviation.
"""


import os
import tensorflow as tf
from tables import *
from dxpy.core.path import Path
from dxpy.configs import configurable
import h5py
import numpy as np
DEFAULT_FILE_NAME = 'analytical_phantom_sinogram.h5'
from ...config import get_config
from ..base import DatasetFromTFDataset, DatasetFromGenerator


class AnalyticalPhantomSinogram(IsDescription):
    phantom = UInt16Col(shape=(256, 256))
    sinogram = UInt16Col(shape=(320, 320))
    phantom_type = UInt8Col()


class AnalyticalPhantomMultiScaleReconstruction(IsDescription):
    recon1x = UInt16Col(shape=(256, 256))
    recon2x = UInt16Col(shape=(256, 256))
    recon4x = UInt16Col(shape=(256, 256))
    recon8x = UInt16Col(shape=(256, 256))


def data_type(field_name):
    import tensorflow as tf
    return {
        'phantom': np.float32,
        'sinogram': np.float32,
        'phantom_type': np.int64
    }[field_name]


def data_type_tf(field_name):
    import tensorflow as tf
    return {
        'phantom': tf.float32,
        'sinogram': tf.float32,
        'phantom_type': tf.int64
    }[field_name]


def data_shape(field_name):
    import tensorflow as tf
    return {
        'phantom': (256, 256),
        'sinogram': (640, 320),
        'phantom_type': (),
    }[field_name]


def h5filename():
    from ..config import config
    return str(Path(config['PATH_DATASETS']) / DEFAULT_FILE_NAME)


def _post_processing(result):
    from dxpy.medical_image_processing.projection.parallel import padding_pi2full
    if 'sinogram' in result:
        result['sinogram'] = padding_pi2full(result['sinogram']).T
    for k in result:
        result[k] = result[k].astype(data_type(k))
    return result


def dataset(fields=('phantom', 'sinogram'), filename=None, idxs=None):
    with open_file(h5filename()) as h5file:
        if idxs is None:
            rows = h5file.root.data.iterrows()
        else:
            rows = idxs
        for d in rows:
            if isinstance(d, int):
                d = h5file.root.data[d]
            result = {k: d[k] for k in fields}
            result = _post_processing(result)
            yield result


def dataset_multi_image(fields=('phantom', 'sinogram', 'recon1x', 'recon2x', 'recon4x', 'recon8x'), filename_sinogram=None, filename_recon=None, idxs=None):
    if filename_recon is None:
        filename_recon = ''
    # TODO: finish
    # 施工中


def _tensorflow_raw_dataset(fields, idxs=None):
    import tensorflow as tf
    from functools import partial
    dataset_gen = partial(dataset, fields=fields, idxs=idxs)
    return tf.data.Dataset.from_generator(dataset_gen, {k: data_type_tf(k) for k in fields}, {k: data_shape(k) for k in fields})


class AnalyticalPhantomSinogramDataset(DatasetFromGenerator):
    @configurable(get_config(), with_name=True)
    def __init__(self, name='dataset', batch_size=32, fields=('sinogram', 'phantom'), idxs=None, shuffle=True):
        from functools import partial
        dataset_gen = partial(dataset, fields=fields, idxs=idxs)
        output_types = {k: data_type_tf(k) for k in fields}
        output_shapes = {k: data_shape(k) for k in fields}
        super().__init__(name, dataset_gen, batch_size=batch_size, idxs=idxs,
                         output_types=output_types, output_shapes=output_shapes, shuffle=shuffle)

    def _reshape_tensors(self, data):
        return {'sinogram': tf.reshape(tf.cast(data['sinogram'], tf.float32),
                                       data_shape('sinogram')),
                'phantom': tf.reshape(tf.cast(data['phantom'], tf.float32),
                                      data_shape('phantom'))}

    def _processing(self):
        dataset = (super()._processing()
                   .repeat()
                   .map(self._reshape_tensors))
        if self.param('shuffle'):
            dataset = dataset.shuffle(1024)
        return dataset.batch(self.param('batch_size'))


@configurable(get_config(), with_name=True)
def get_dataset(name='dataset', fields=['sinogram']):
    from ..base import DatasetFromTFDataset
    return DatasetFromTFDataset(name, _tensorflow_raw_dataset(fields))

import typing
@configurable(get_config(), with_name=True)
def analytical_super_resolution_dataset(image_type: str,
                                        poission_noise: bool,
                                        batch_size: int,
                                        nb_down_sample: int,
                                        target_shape: typing.List[int],
                                        idxs: typing.List[int],
                                        *,
                                        name: str='dataset',
                                        path_dataset: "str|None"=None):
    """
    Args:
        -   image_type: 'sinogram' or 'image'
        -   batch_size
    Returns:
        a `Graph` object, which has several nodes:
    Raises:
    """
    from ...model.normalizer.normalizer import FixWhite, ReduceSum
    from ..super_resolution import SuperResolutionDataset
    from ...model.tensor import ShapeEnsurer
    from ...config import config
    normalizer_configs = {
        'analytical_phantoms': {'mean': 4.88, 'std': 4.68}
    }
    config_origin = {}
    config_normalizer = normalizer_configs[dataset_name]
    config['dataset'] = {
        'origin': config_origin,
        'fix_white': config_normalizer
    }
    with tf.name_scope('{img_type}_dataset'.format(img_type=image_type)):
        if image_type == 'sinogram':
            dataset_origin = AnalyticalPhantomSinogramDataset(name=name, batch_size=batch_size, fields=['sinogram'], idxs)
        dataset_summed = ReduceSum('dataset/reduce_sum',
                                   dataset_origin[image_type],
                                   fixed_summation_value=1e6).as_tensor()

        dataset = FixWhite(name='dataset/fix_white',
                           inputs=dataset_summed)()
        if poission_noise:
            with tf.name_scope('add_poission_noise'):
                dataset = tf.random_poisson(dataset, shape=[])
        dataset = tf.random_crop(dataset,
                                 [batch_size] + list(target_shape) + [1])
        dataset = SuperResolutionDataset('dataset/super_resolution',
                                         lambda: {'image': dataset},
                                         input_key='image',
                                         nb_down_sample=3)
    return dataset




# The following part is a record of dataset generator script.


# def create_dataset_from_tf_records():
#     from tqdm import tqdm
#     from dxpy.learn.dataset.config import config
#     filename = '/media/hongxwing/73f574f4-33b0-46b1-ae5d-de1eca4ef891/home/hongxwing/Workspace/Datas/phantom_sinograms_3.h5'
#     with h5py.File(filename, 'r') as fin:
#         filters = Filters(5, 'blosc')
#         h5file = open_file(str(Path(config['PATH_DATASETS']) / DEFAULT_FILE_NAME),
#                            mode="w", filters=filters)
#         table = h5file.create_table(
#             h5file.root, 'data', AnalyticalPhantomSinogram, expectedrows=fin['phantoms'].shape[0])
#         data = table.row

#         def convert_to_uint16(arr):
#             UINT16MAX = 65535
#             arr = np.maximum(arr, 0.0)
#             arr = arr / np.max(arr) * UINT16MAX
#             arr = np.minimum(arr, UINT16MAX)
#             return arr.astype(np.uint32)

#         for i in tqdm(range(fin['phantoms'].shape[0]), ascii=True):
#             data['sinogram'] = convert_to_uint16(
#                 fin['sinograms'][i, :, :320, 0])
#             data['phantom'] = convert_to_uint16(fin['phantoms'][i, :, :, 0])
#             data['phantom_type'] = np.reshape(np.array(fin['types'][i, ...],
#                                                        dtype=np.uint8), [])
#             data.append()
#     table.flush()
#     h5file.close()


# if __name__ == "__main__":
#     create_dataset_from_tf_records()
